"""
pygame-menu
https://github.com/ppizarror/pygame-menu

TABLE
The table widget is a Frame which packs widgets in a structured way.
"""

__all__ = [
    'Table',
    'TableManager'
]

import pygame
import pygame_menu

from abc import ABC
from pygame_menu.baseimage import BaseImage
from pygame_menu.font import FontType, assert_font
from pygame_menu.locals import ORIENTATION_VERTICAL, ALIGN_LEFT, ALIGN_CENTER, \
    ORIENTATION_HORIZONTAL, POSITION_NORTH, POSITION_CENTER, POSITION_SOUTH, \
    ALIGN_RIGHT, POSITION_WEST, POSITION_EAST
from pygame_menu.utils import assert_alignment, assert_color, uuid4, parse_padding, \
    assert_position, assert_vector, warn
from pygame_menu.version import ver
from pygame_menu.widgets.core.widget import Widget, WidgetBorderPositionType, \
    WIDGET_FULL_BORDER, WIDGET_BORDER_POSITION_NONE, AbstractWidgetManager
from pygame_menu.widgets.widget.frame import Frame
from pygame_menu.widgets.widget.image import Image
from pygame_menu.widgets.widget.label import Label
from pygame_menu.widgets.widget.surface import SurfaceWidget

from pygame_menu._types import List, Union, ColorInputType, Optional, Tuple, \
    VectorInstance, PaddingType, Dict, NumberType, Vector2IntType, EventVectorType

CellType = Union['Widget', str, int, float, bool, 'BaseImage', 'pygame.Surface']
ColumnInputType = Union[Tuple[CellType, ...], List[CellType]]


# noinspection PyMissingOrEmptyDocstring
class Table(Frame):
    """
    Table is a frame which can pack widgets in a structured way.

    .. note::

        Table cannot be selected. Thus, it does not receive any selection effect.

    .. note::

        Table only accepts translation and resize transformations.

    :param table_id: ID of the table
    """
    _rows: List['Frame']
    _update_widgets: List['Widget']
    default_cell_align: str
    default_cell_border_color: ColorInputType
    default_cell_border_position: WidgetBorderPositionType
    default_cell_border_width: int
    default_cell_padding: PaddingType
    default_cell_vertical_position: str
    default_row_background_color: Optional[ColorInputType]

    def __init__(
        self,
        table_id: str = ''
    ) -> None:
        super(Table, self).__init__(
            width=1,
            height=1,
            orientation=ORIENTATION_VERTICAL,
            frame_id=table_id
        )

        # Internals
        self._rows = []
        self._update_widgets = []

        # Frame behaviour
        self._accepts_scrollarea = False
        self._accepts_title = False

        # Set size
        self._frame_size = (0, 0)
        self._height = 0
        self._real_rect = pygame.Rect(0, 0, 0, 0)
        self._width = 0

        # Default cell properties
        self.default_cell_align = ALIGN_LEFT
        self.default_cell_border_color = (0, 0, 0)
        self.default_cell_border_position = WIDGET_FULL_BORDER
        self.default_cell_border_width = 1
        self.default_cell_padding = 0
        self.default_cell_vertical_position = POSITION_NORTH
        self.default_row_background_color = None

        # Finals
        self.relax()

    def pack(self, *args, **kwargs) -> None:
        raise RuntimeError(f'{self.get_class_id()} cannot pack external widgets')

    def remove_row(self, row: 'Frame') -> None:
        """
        Removes row from the table.

        :param row: Row frame
        """
        self.unpack(row)

    def unpack(self, row: 'Frame') -> None:
        assert row != self, 'table cannot unpack itself'
        assert len(self._widgets) > 0, 'table is empty'
        assert row in self._rows and row.get_id() in self._widgets.keys(), \
            f'row {row.get_class_id()} does not exist on {self.get_class_id()}'
        wid = row.get_id()
        assert row._frame == self, 'widget frame differs from current'
        row._frame = None
        row._translate_virtual = (0, 0)
        del self._widgets[wid]
        try:
            del self._pos[wid]
        except KeyError:
            pass
        self._rows.remove(row)
        self._menu_render()
        self._update_row_sizing()
        self._update_event_widgets()

        # Remove scrollable from rows
        if self._menu is not None:
            total_removed = 0
            menu_update_frames = self._get_menu_update_frames()
            for w in row.get_widgets(unpack_subframes=False):
                if isinstance(w, Frame):
                    if w in menu_update_frames:
                        menu_update_frames.remove(w)
                        total_removed += 1
            if total_removed > 0:
                self._sort_menu_update_frames()

    @staticmethod
    def _check_cell_style(
        align: str,
        background_color: ColorInputType,
        border_color: ColorInputType,
        border_position: WidgetBorderPositionType,
        border_width: int,
        padding: PaddingType,
        vertical_position: str
    ) -> None:
        """
        Assert cell style.

        :param align: Horizontal align of each cell. See :py:mod:`pygame_menu.locals`
        :param background_color: Background color
        :param border_color: Border color of each cell
        :param border_position: Border position of each cell. Valid only: north, south, east, and west. See :py:mod:`pygame_menu.locals`
        :param border_width: Border width in px of each cell
        :param padding: Cell padding according to CSS rules. General shape: (top, right, bottom, left)
        :param vertical_position: Vertical position of each cell. Only valid: north, center, and south. See :py:mod:`pygame_menu.locals`
        """
        # Alignment
        assert_alignment(align)

        # Background color
        if background_color is not None:
            assert_color(background_color)

        # Padding
        parse_padding(padding)

        # Vertical position
        assert_position(vertical_position)
        assert vertical_position in (POSITION_NORTH, POSITION_CENTER, POSITION_SOUTH), \
            'cell vertical position must be NORTH, CENTER, or SOUTH'

        # Border color
        assert isinstance(border_width, int) and border_width >= 0
        if border_color is not None:
            assert_color(border_color)

        # Border position
        assert isinstance(border_position, (str, VectorInstance))
        if isinstance(border_position, str):
            border_position = [border_position]

        # Border positioning
        for pos in border_position:
            assert pos in (POSITION_NORTH, POSITION_SOUTH, POSITION_EAST, POSITION_WEST), \
                f'only north, south, east, and west border positions are valid, ' \
                f'but received "{pos}"'

    def add_row(
        self,
        cells: Union[ColumnInputType, 'Widget'],
        cell_align: Optional[str] = None,
        cell_border_color: Optional[ColorInputType] = None,
        cell_border_position: Optional[WidgetBorderPositionType] = None,
        cell_border_width: Optional[int] = None,
        cell_font: Optional[FontType] = None,
        cell_font_color: Optional[ColorInputType] = None,
        cell_font_size: Optional[int] = None,
        cell_padding: PaddingType = None,
        cell_vertical_position: Optional[str] = None,
        row_background_color: Optional[ColorInputType] = None
    ) -> 'Frame':
        """
        Add row to table.

        .. note::

            By default, if ``None`` each cell style uses the table defaults "cell"
            styles.

        .. note::

            By default, the cell font is the same as the table font style.

        .. warning::

            Currently, only static widgets work properly, that is, widgets that
            does not accept events. Buttons or TextInputs can be added, but them
            will not accept any event, also, these widgets cannot be selected.
            That's because the Table architecture relies on Frame widget.

        :param cells: Cells to add. This can be a tuple or list of widgets, string, numbers, boolean values or images. Also, a Frame row can be added
        :param cell_align: Horizontal align of each cell. See :py:mod:`pygame_menu.locals`
        :param cell_border_color: Border color of each cell
        :param cell_border_position: Border position of each cell. Valid only: north, south, east, and west. See :py:mod:`pygame_menu.locals`
        :param cell_border_width: Border width in px of each cell
        :param cell_font: Font name or path
        :param cell_font_color: Font color
        :param cell_font_size: Font size
        :param cell_padding: Padding of each cell according to CSS rules. General shape: (top, right, bottom, left)
        :param cell_vertical_position: Vertical position of each cell. Only valid: north, center, and south. See :py:mod:`pygame_menu.locals`
        :param row_background_color: Row background color
        :return:
        """
        assert self.configured, 'table must be configured before adding rows'

        # Use defaults
        if cell_align is None:
            cell_align = self.default_cell_align
        if cell_border_color is None:
            cell_border_color = self.default_cell_border_color
        if cell_border_position is None:
            cell_border_position = self.default_cell_border_position
        if cell_border_width is None:
            cell_border_width = self.default_cell_border_width
        if cell_font is None:
            cell_font = self._font_name
        if cell_font_color is None:
            cell_font_color = self._font_color
        if cell_font_size is None:
            cell_font_size = self._font_size
        if cell_padding is None:
            cell_padding = self.default_cell_padding
        if cell_vertical_position is None:
            cell_vertical_position = self.default_cell_vertical_position
        if row_background_color is None:
            row_background_color = self.default_row_background_color

        # If cells is a previous table row
        if isinstance(cells, Frame) and cells.has_attribute('is_row'):
            _row_cells = list(cells.get_widgets(unpack_subframes=False))
            cells.clear()
            cells = _row_cells
        if isinstance(cells, Widget):
            cells = [cells]

        assert isinstance(cells, VectorInstance)

        # Check cell styles
        self._check_cell_style(
            align=cell_align,
            background_color=row_background_color,
            border_color=cell_border_color,
            border_position=cell_border_position,
            border_width=cell_border_width,
            padding=cell_padding,
            vertical_position=cell_vertical_position
        )
        cell_padding = parse_padding(cell_padding)
        if cell_border_color is not None:
            cell_border_color = assert_color(cell_border_color)
        if isinstance(cell_border_position, str):
            cell_border_position = [cell_border_position]

        # Check positioning
        if cell_border_width == 0:
            cell_border_position = WIDGET_BORDER_POSITION_NONE

        if row_background_color is not None:
            row_background_color = assert_color(row_background_color)

        # Create frame row
        row = Frame(1, 1, ORIENTATION_HORIZONTAL, frame_id=self._id + '+cell-row-' + uuid4(short=True))
        row._accepts_scrollarea = False
        row._accepts_title = False
        row._menu_can_be_none_pack = True
        row._update__repr___(self)
        row.configured = True
        row.relax()
        row.set_background_color(row_background_color)
        row.set_menu(self._menu)
        row.set_scrollarea(self._scrollarea)
        row.set_attribute('is_row')
        row.set_controls(
            joystick=self._joystick_enabled,
            mouse=self._mouse_enabled,
            touchscreen=self._touchscreen_enabled,
            keyboard=self._keyboard_enabled
        )
        # row.set_frame(self) This cannot be executed as row is packed within

        # Create widgets
        row_cells: List['Widget'] = []
        cell: 'Widget'
        j = 0

        for c in cells:
            cell_widget_type = False

            if isinstance(c, (str, int, float, bool)):
                cell = Label(c, label_id=self._id + '+cell-label-' + uuid4(short=True))
                cell.set_font(
                    antialias=self._font_antialias,
                    background_color=None,
                    color=cell_font_color,
                    font=cell_font,
                    font_size=cell_font_size,
                    readonly_color=self._font_readonly_color,
                    readonly_selected_color=self._font_readonly_selected_color,
                    selected_color=self._font_selected_color
                )
                cell.set_padding(0)
                cell.set_tab_size(self._tab_size)

            elif isinstance(c, BaseImage):
                cell = Image(
                    c, image_id=self._id + '+cell-image-' + uuid4(short=True)
                )

            elif isinstance(c, pygame.Surface):
                cell = SurfaceWidget(
                    c, surface_id=self._id + '+cell-surface-' + uuid4(short=True)
                )

            else:
                assert isinstance(c, Widget)
                assert c != self, f'{self.get_class_id()} cannot be appended to itself'

                # Check if Frame not recursive
                if isinstance(c, Frame):
                    assert self not in c.get_widgets(unpack_subframes_include_frame=True), \
                        f'{self.get_class_id()} cannot be packed within {c.get_class_id()},' \
                        f' recursive packing is not allowed (Table is within Frame' \
                        f' to be inserted as row cell)'

                cell = c
                if c._accept_events:
                    cell_widget_type = True
                    if self._verbose:
                        warn(
                            f'{self.get_class_id()} does not accept events in current'
                            f' pygame-menu v{ver}; thus appended cell row widget '
                            f'{c.get_class_id()} (pos {j}) would not work properly, '
                            f'as it will ignore all inputs. Also, widgets within Tables'
                            f' cannot be selected. Consider Tables as visual-only'
                        )
                # self._append_menu_update_frame(self)

            # Configure cell
            cell.set_attribute('accept_events', cell_widget_type)
            cell.set_attribute('align', cell_align)
            cell.set_attribute('background_color', row_background_color)
            cell.set_attribute('border_color', cell_border_color)
            cell.set_attribute('border_position', cell_border_position)
            cell.set_attribute('border_width', cell_border_width)
            cell.set_attribute('column', j + 1)
            cell.set_attribute('padding', cell_padding)
            cell.set_attribute('row', len(self._rows) + 1)
            cell.set_attribute('row_frame', row)
            cell.set_attribute('table', self)
            cell.set_attribute('vertical_position', cell_vertical_position)
            cell.set_float(False)
            cell._update__repr___(self)
            cell.configured = True

            # If cell is within a menu, remove from it
            if cell.get_menu() is not None:
                try:
                    cell.get_menu().remove_widget(cell)
                except ValueError:
                    pass

            # Check the cell frame is None
            assert cell.get_frame() != self, \
                f'{cell.get_class_id()} cannot be added as it already exists in table'
            assert cell.get_frame() is None, \
                f'{cell.get_class_id()} is already packed in ' \
                f'{cell.get_frame().get_class_id()}, it cannot be added to {self.get_class_id()}'

            # If cell is frame and scrollable
            if isinstance(cell, Frame):
                self._append_menu_update_frame(cell)

            # Add to cells
            row_cells.append(cell)
            j += 1

        # Pack cells to row
        for c in row_cells:
            row.pack(c)

        # Pack rows to self
        super(Table, self).pack(row)
        self._rows.append(row)

        # Update size
        self._update_row_sizing()
        self._update_event_widgets()

        return row

    def _update_event_widgets(self) -> None:
        """
        Update the list of widgets that accept events.
        """
        self._update_widgets = []
        for r in self._rows:
            for w in r.get_widgets():
                if w.get_attribute('accept_events'):
                    self._update_widgets.append(w)

    def _get_column_width_row_height(self) -> Tuple[Dict[int, int], Dict['Frame', int]]:
        """
        Return column width and row height.

        :return: Column width and row height dict
        """
        column_widths: Dict[int, int] = {}  # column/width
        row_heights: Dict['Frame', int] = {}  # row/height

        for f in self._rows:
            col = 0  # Column
            max_height = 0  # Max row height
            for w in f.get_widgets(unpack_subframes=False):
                width = w.get_width(apply_padding=False)
                # Add inner padding
                pad = w.get_attribute('padding')  # top, right, bottom, left
                width += pad[1] + pad[3]
                if col not in column_widths.keys():
                    column_widths[col] = width
                else:
                    column_widths[col] = max(width, column_widths[col])
                col += 1
                height = w.get_height(apply_padding=False)
                height += pad[0] + pad[2]
                max_height = max(max_height, height)
            row_heights[f] = max_height

        return column_widths, row_heights

    # def set_frame(self, frame: 'pygame_menu.widgets.Frame') -> 'Table':
    #     super(Frame, self).set_frame(frame)
    #     for f in self._rows:
    #         f.set_frame(frame)
    #     return self

    def _update_row_sizing(self) -> None:
        """
        Update row sizing.
        """
        column_widths, row_heights = self._get_column_width_row_height()

        total_width = 0
        # Sum each column to get max row width
        for w in column_widths.values():
            total_width += w

        # Compute each row height and update padding
        total_height = 0
        for f in self._rows:
            f.resize(total_width, row_heights[f])
            total_height += row_heights[f]

            col = 0
            for w in f.get_widgets(unpack_subframes=False):
                w_w = w.get_width(apply_padding=False)
                w_h = w.get_height(apply_padding=False)
                w_pad = w.get_attribute('padding')  # top, right, bottom, left
                w_align = w.get_attribute('align')
                w_vpos = w.get_attribute('vertical_position')
                w_total_height = row_heights[f]
                w_total_width = column_widths[col]

                # Default paddings
                pad_top = w_pad[0]
                pad_right = w_pad[1]
                pad_bottom = w_pad[2]
                pad_left = w_pad[3]

                # Subtract padding to max width/height
                w_total_height -= (pad_top + pad_bottom)
                w_total_width -= (pad_left + pad_right)

                # Compute horizontal align
                delta_w = w_total_width - w_w
                assert delta_w >= 0, 'delta width cannot be negative'
                if w_align == ALIGN_LEFT:
                    pad_right += delta_w
                elif w_align == ALIGN_CENTER:
                    pad_left += int(delta_w / 2)
                    pad_right += int(delta_w / 2)
                elif w_align == ALIGN_RIGHT:
                    pad_left += delta_w

                # Compute vertical position
                delta_h = w_total_height - w_h
                assert delta_h >= 0, 'delta height cannot be negative'
                if w_vpos == POSITION_NORTH:
                    pad_bottom += delta_h
                elif w_vpos == POSITION_CENTER:
                    pad_top += int(delta_h / 2)
                    pad_bottom += int(delta_h / 2)
                elif w_vpos == POSITION_SOUTH:
                    pad_top += delta_h

                # Update cell padding
                w.set_padding((pad_top, pad_right, pad_bottom, pad_left))

                # Check if the padding closes the width
                dx = column_widths[col] - w.get_width()
                dy = row_heights[f] - w.get_height()
                if dx == 1:
                    pad_left += 1
                if dx >= 2:
                    pad_right += (dx - 1)
                if dy == 1:
                    pad_top += 1
                if dy >= 2:
                    pad_bottom += (dy - 1)
                w.set_padding((pad_top, pad_right, pad_bottom, pad_left))

                col += 1

        # Update current rect
        self.resize(total_width + self._padding[1] + self._padding[3],
                    total_height + self._padding[0] + self._padding[2])

    def on_remove_from_menu(self) -> 'Frame':
        self.update_indices()
        return self

    @staticmethod
    def get_cell_column_row(cell: 'Widget') -> Tuple[int, int]:
        """
        Return the column/row within table layout for the given widget.

        :param cell: Widget (cell) to get the column/row position
        :return: Column/Row
        """
        assert cell.has_attribute('column') and cell.has_attribute('row'), \
            f'{cell.get_class_id()} does not have the table attributes'
        return cell.get_attribute('column'), cell.get_attribute('row')

    def _draw_cell_borders(self, surface: 'pygame.Surface') -> None:
        """
        Draw cell border.

        :param surface: Surface to draw the cell border
        """
        if len(self._rows) == 0:
            return

        # Get first row position
        x, y = self._rows[0].get_position()

        column_widths, row_heights = self._get_column_width_row_height()
        total_height = 0
        r = 0
        for row in self._rows:
            r += 1
            col = 0
            total_width = 0
            for w in row.get_widgets(unpack_subframes=False):
                border_color = w.get_attribute('border_color')
                border_position = w.get_attribute('border_position')
                border_width = w.get_attribute('border_width')

                # Create drawing rect
                subtract_border = (-border_width) if r == len(self._rows) else 0
                rect = pygame.Rect(total_width + x,
                                   total_height + y,
                                   column_widths[col],
                                   row_heights[row] + subtract_border)
                total_width += column_widths[col]
                col += 1

                # Draw the border
                if border_position == WIDGET_BORDER_POSITION_NONE or border_width == 0:
                    continue
                for pos in border_position:
                    if pos == POSITION_NORTH:
                        start, end = rect.topleft, rect.topright
                    elif pos == POSITION_SOUTH:
                        start, end = rect.bottomleft, rect.bottomright
                    elif pos == POSITION_EAST:
                        start, end = rect.topright, rect.bottomright
                    elif pos == POSITION_WEST:
                        start, end = rect.topleft, rect.bottomleft
                    else:
                        raise RuntimeError(f'invalid border position "{pos}"')
                    pygame.draw.line(
                        surface,
                        border_color,
                        start,
                        end,
                        border_width
                    )

            total_height += row_heights[row]

    def get_cell(self, column: int, row: int) -> 'Widget':
        """
        Get cell widget at column/row position.

        :param column: Cell column position (counting from 1)
        :param row: Cell row position (counting from 1)
        :return: Cell widget object
        """
        assert isinstance(row, int) and row >= 1, \
            'row index must be an integer equal or greater than 1'
        assert isinstance(column, int) and column >= 1, \
            'column index must be an integer equal or greater than 1'
        assert row <= len(self._rows), \
            f'row index ({row}) cannot exceed the number of rows ({len(self._rows)})'
        f = self._rows[row - 1]
        w = f.get_widgets(unpack_subframes=False)
        assert column <= len(w), \
            f'column index ({column}) cannot exceed the number of columns ({len(w)}) of row {row}'
        return w[column - 1]

    def is_rectangular(self) -> bool:
        """
        Return ``True`` if the table is rectangular, that is, each row have the
        same number of columns.

        :return: Bool
        """
        if len(self._rows) == 0:
            return True
        c = self._rows[0].get_total_packed()
        for f in self._rows:
            if f.get_total_packed() != c:
                return False
        return True

    def update_cell_style(
        self,
        column: Union[int, Vector2IntType],
        row: Union[int, Vector2IntType],
        align: Optional[str] = None,
        background_color: Optional[ColorInputType] = None,
        border_color: Optional[ColorInputType] = None,
        border_position: Optional[WidgetBorderPositionType] = None,
        border_width: Optional[int] = None,
        font: Optional[FontType] = None,
        font_color: Optional[ColorInputType] = None,
        font_size: Optional[int] = None,
        padding: Optional[PaddingType] = None,
        vertical_position: Optional[str] = None
    ) -> Union['Widget', List['Widget']]:
        """
        Update cell style. If a parameter is ``None`` the default cell property
        will be used.

        :param column: Cell column position (counting from 1). If -1 update all column from the given row. Also, a 2-item list/tuple is accepted (from, to), ``to=-1`` is also accepted (last)
        :param row: Cell row position (counting from 1). If ``-1`` update all rows from the given column. Also, a 2-item list/tuple is accepted (from, to), ``to=-1`` is also accepted (last)
        :param align: Horizontal align of each cell. See :py:mod:`pygame_menu.locals`
        :param background_color: Background color
        :param border_color: Border color of each cell
        :param border_position: Border position of each cell. Valid only: north, south, east, and west. See :py:mod:`pygame_menu.locals`
        :param border_width: Border width in px of each cell
        :param font: Font name or path
        :param font_color: Font color
        :param font_size: Font size
        :param padding: Cell padding according to CSS rules. General shape: (top, right, bottom, left)
        :param vertical_position: Vertical position of each cell. Only valid: north, center, and south. See :py:mod:`pygame_menu.locals`
        :return: Cell widget
        """
        if row == -1 or isinstance(row, VectorInstance):
            max_rows = len(self._rows)
            if row == -1:
                row = []
                for i in range(max_rows):
                    row.append(i + 1)
            else:
                assert_vector(row, 2, int)
                row_k = list(row)
                if row_k[1] == -1:
                    row_k[1] = len(self._rows)
                assert 1 <= row_k[0] <= row_k[1] <= max_rows, \
                    f'(from, to) of rows vector must be increasing and between 1-{max_rows}'
                row = [row_k[0]]
                for i in range(row_k[1] - row_k[0]):
                    row.append(row_k[0] + (i + 1))
            if isinstance(column, VectorInstance) and column != [1, -1]:
                assert self.is_rectangular(), \
                    f'only rectangular tables (same number of columns for each row) ' \
                    f'accept a variable column different than -1 or [1, -1], but ' \
                    f'received "{column}"'
            updated_wid = []
            for i in row:
                w = self.update_cell_style(
                    column=column,
                    row=i,
                    align=align,
                    background_color=background_color,
                    border_color=border_color,
                    border_position=border_position,
                    border_width=border_width,
                    font=font,
                    font_color=font_color,
                    font_size=font_size,
                    padding=padding,
                    vertical_position=vertical_position
                )
                if not isinstance(w, list):
                    w = [w]
                for k in w:
                    updated_wid.append(k)
            return updated_wid
        if column == -1 or isinstance(column, VectorInstance):
            assert isinstance(row, int) and 1 <= row <= len(self._rows), \
                f'row index ({row}) cannot exceed the number of rows ({len(self._rows)})'
            max_columns = self._rows[row - 1].get_total_packed()
            if column == -1:
                column = []
                for i in range(max_columns):
                    column.append(i + 1)
            else:
                assert_vector(column, 2, int)
                column_k = list(column)
                if column_k[1] == -1:
                    column_k[1] = max_columns
                assert 1 <= column_k[0] <= column_k[1] <= max_columns, \
                    f'(from, to) of column vector must be increasing and between 1-{max_columns} for row {row}'
                column = [column_k[0]]
                for i in range(column_k[1] - column_k[0]):
                    column.append(column_k[0] + (i + 1))
            updated_wid = []
            for i in column:
                w = self.update_cell_style(
                    column=i,
                    row=row,
                    align=align,
                    background_color=background_color,
                    border_color=border_color,
                    border_position=border_position,
                    border_width=border_width,
                    font=font,
                    font_color=font_color,
                    font_size=font_size,
                    padding=padding,
                    vertical_position=vertical_position
                )
                if not isinstance(w, list):
                    w = [w]
                for k in w:
                    updated_wid.append(k)
            return updated_wid
        cell = self.get_cell(column, row)
        r = self._rows[row - 1]

        if align is None:
            align = cell.get_attribute('align')
        if background_color is None:
            background_color = cell.get_attribute('background_color')
        if border_color is None:
            border_color = cell.get_attribute('border_color')
        if border_position is None:
            border_position = cell.get_attribute('border_position')
        if border_width is None:
            border_width = cell.get_attribute('border_width')
        if padding is None:
            padding = cell.get_attribute('padding')
        if vertical_position is None:
            vertical_position = cell.get_attribute('vertical_position')

        self._check_cell_style(
            align=align,
            background_color=background_color,
            border_color=border_color,
            border_position=border_position,
            border_width=border_width,
            padding=padding,
            vertical_position=vertical_position
        )
        if background_color is not None:
            background_color = assert_color(background_color)
        if border_color is not None:
            border_color = assert_color(border_color)
        padding = parse_padding(padding)

        # Update background color
        if background_color != r._background_color:
            cell.set_background_color(background_color)
        else:
            cell.set_background_color(None)

        # Update font
        if font_color is None:
            font_color = cell._font_color
        assert_color(font_color)
        if font is None:
            font = cell._font_name
        assert_font(font)
        if font_size is None:
            font_size = cell._font_size

        try:
            cell.update_font({
                'color': font_color,
                'name': font
            })
        except AssertionError:
            pass

        try:
            if isinstance(font_size, int) and font_size > 0:
                cell.update_font({'size': font_size})
        except AssertionError:
            pass

        if isinstance(border_position, str):
            border_position = [border_position]

        # Update cell
        cell.set_attribute('align', align)
        cell.set_attribute('background_color', background_color)
        cell.set_attribute('border_color', border_color)
        cell.set_attribute('border_position', border_position)
        cell.set_attribute('border_width', border_width)
        cell.set_attribute('padding', padding)
        cell.set_attribute('vertical_position', vertical_position)

        self._update_row_sizing()
        self._render()
        self.force_menu_surface_update()
        return cell

    # noinspection PyProtectedMember
    def set_scrollarea(self, scrollarea: Optional['pygame_menu._scrollarea.ScrollArea']) -> None:
        super(Table, self).set_scrollarea(scrollarea)
        for f in self._rows:
            f.set_scrollarea(scrollarea)

    def set_position(self, x: NumberType, y: NumberType) -> 'Table':
        super(Table, self).set_position(x, y)
        x = self._rect.x
        y = self._rect.y
        for f in self._rows:
            f.set_position(x, y)
            for w in f.get_widgets(unpack_subframes=False):
                w._set_position_relative_to_frame()
            f.update_position()
        return self

    def draw(self, surface: 'pygame.Surface') -> 'Table':
        if not self.is_visible():
            return self
        super(Table, self).draw(surface)
        self._draw_cell_borders(self.last_surface)

    def update(self, events: EventVectorType) -> bool:
        super(Table, self).update(events)
        updated = False

        # if self.readonly or not self.is_visible():
        #     return updated
        #
        # for w in self._update_widgets:
        #     updated = w.update(events) or updated

        return updated


class TableManager(AbstractWidgetManager, ABC):
    """
    Table manager.
    """

    def table(
        self,
        table_id: str = '',
        **kwargs
    ) -> 'pygame_menu.widgets.Table':
        """
        Adds a Table to the Menu. A table is a frame which can pack widgets in a
        structured way.

        kwargs (Optional)
            - ``align``                         (str) – Widget `alignment <https://pygame-menu.readthedocs.io/en/latest/_source/themes.html#alignment>`_
            - ``background_color``              (tuple, list, str, int, :py:class:`pygame.Color`, :py:class:`pygame_menu.baseimage.BaseImage`) – Color of the background. ``None`` for no-color
            - ``background_inflate``            (tuple, list) – Inflate background on x-axis and y-axis (x, y) in px
            - ``border_color``                  (tuple, list, str, int, :py:class:`pygame.Color`) – Widget border color. ``None`` for no-color
            - ``border_inflate``                (tuple, list) – Widget border inflate on x-axis and y-axis (x, y) in px
            - ``border_position``               (str, tuple, list) – Widget border positioning. It can be a single position, or a tuple/list of positions. Only are accepted: north, south, east, and west. See :py:mod:`pygame_menu.locals`
            - ``border_width``                  (int) – Border width in px. If ``0`` disables the border
            - ``cursor``                        (int, :py:class:`pygame.cursors.Cursor`, None) – Cursor of the frame if the mouse is placed over
            - ``float``                         (bool) - If ``True`` the widget don't contribute width/height to the Menu widget positioning computation, and don't add one unit to the rows
            - ``float_origin_position``         (bool) - If ``True`` the widget position is set to the top-left position of the Menu if the widget is floating
            - ``font_background_color``         (tuple, list, str, int, :py:class:`pygame.Color`, None) – Widget font background color
            - ``font_color``                    (tuple, list, str, int, :py:class:`pygame.Color`) – Widget font color
            - ``font_name``                     (str, :py:class:`pathlib.Path`, :py:class:`pygame.font.Font`) – Widget font path
            - ``font_shadow_color``             (tuple, list, str, int, :py:class:`pygame.Color`) – Font shadow color
            - ``font_shadow_offset``            (int) – Font shadow offset in px
            - ``font_shadow_position``          (str) – Font shadow position, see locals for position
            - ``font_shadow``                   (bool) – Font shadow is enabled or disabled
            - ``font_size``                     (int) – Font size of the widget
            - ``margin``                        (tuple, list) – Widget (left, bottom) margin in px
            - ``max_height``                    (int) – Max height in px. If lower than the frame height a scrollbar will appear on vertical axis. ``None`` by default (same height)
            - ``max_width``                     (int) – Max width in px. If lower than the frame width a scrollbar will appear on horizontal axis. ``None`` by default (same width)
            - ``padding``                       (int, float, tuple, list) – Widget padding according to CSS rules. General shape: (top, right, bottom, left)
            - ``shadow_color``                  (tuple, list, str, int, :py:class:`pygame.Color`) – Color of the widget shadow
            - ``shadow_radius``                 (int) - Border radius of the shadow
            - ``shadow_type``                   (str) - Shadow type, it can be ``'rectangular'`` or ``'ellipse'``
            - ``shadow_width``                  (int) - Width of the shadow. If ``0`` the shadow is disabled

        .. note::

            All theme-related optional kwargs use the default Menu theme if not
            defined.

        .. note::

            This is applied only to the base Menu (not the currently displayed,
            stored in ``_current`` pointer); for such behaviour apply to
            :py:meth:`pygame_menu.menu.Menu.get_current` object.

        :param table_id: ID of the table
        :param kwargs: Optional keyword arguments
        :return: Widget object
        :rtype: :py:class:`pygame_menu.widgets.Table`
        """
        attributes = self._filter_widget_attributes(kwargs)

        widget = Table(
            table_id=table_id
        )

        self._configure_widget(widget=widget, **attributes)
        self._append_widget(widget)
        self._check_kwargs(kwargs)

        return widget



"""
pygame-menu
https://github.com/ppizarror/pygame-menu

EXAMPLE - WIDGET POSITIONING
Test widget positioning example.
"""

import pygame_menu
from pygame_menu.examples import create_example_window

# Create the surface
surface = create_example_window('Example - Widget Positioning', (640, 480))

# Create a custom theme
my_theme = pygame_menu.themes.THEME_DARK.copy()
my_theme.title = False  # Hide the menu title

menu = pygame_menu.Menu(
    height=480,  # Use full-screen
    theme=my_theme,
    title='',
    center_content=False,
    width=640
)

menu.add.label(
    'My App',
    background_color='#333',
    background_inflate=(30, 0),
    float=True  # Widget does not add size to the menu
).translate(0, 10)

label = menu.add.label(
    'Lorem ipsum',
    float=True,
    font_name=pygame_menu.font.FONT_OPEN_SANS_ITALIC,
    font_size=25)
label.rotate(90)
label.translate(300, 160)

# Button options
b1 = menu.add.button(
    'Main Menu',
    lambda: print(f'My method'),
    align=pygame_menu.locals.ALIGN_LEFT,
    float=True,
    selection_color='#fff'
)
b1.translate(10, 170)
b2 = menu.add.button(
    'Exit',
    pygame_menu.events.EXIT,
    align=pygame_menu.locals.ALIGN_LEFT,
    float=True,
    selection_color='#fff'
)
b2.translate(10, 235)

# Bottom scrollable text
f = menu.add.frame_v(
    background_color='#6b6e5e',
    border_color='#36372f',
    border_width=1,
    float=True,
    height=480,
    max_height=100,
    width=200
)
f.translate(220, 390)
labels = [menu.add.label(f'  Lorem ipsum #{i}', font_size=15, font_color='#000000', padding=0) for i in range(20)]
for j in labels:
    f.pack(j)

if __name__ == '__main__':
    menu.mainloop(surface)



"""
pygame-menu
https://github.com/ppizarror/pygame-menu

PYGAME-MENU
A menu for pygame. Simple, and easy to use.
"""

__all__ = [

    # Common classes
    'BaseImage',
    'Menu',
    'Sound',
    'Theme'

]

# Check if pygame exists, if not maybe the module is being used by setup.py
__pygame_version__ = None
try:
    from pygame import version as __pygame_version__

    __pygame_version__ = __pygame_version__.vernum
except (ModuleNotFoundError, ImportError):
    pass

# Import modules that require pygame
if __pygame_version__ is not None:
    """
    BaseImage: Provides basic image loading and manipulation with pygame
    """
    import pygame_menu.baseimage
    from pygame_menu.baseimage import BaseImage

    """
    Controls: Default controls of menu object and key definition
    """
    import pygame_menu.controls

    """
    Events: Menu events definition and locals
    """
    import pygame_menu.events

    """
    Fonts: Menu fonts
    """
    import pygame_menu.font

    """
    Locals: Local constants
    """
    import pygame_menu.locals

    """
    Menu: Menu class
    """
    from pygame_menu.menu import Menu

    """
    ScrollArea: Scrollarea class
    """
    import pygame_menu._scrollarea

    """
    Sound: Sound class
    """
    import pygame_menu.sound
    from pygame_menu.sound import Sound

    """
    Themes: Menu themes
    """
    import pygame_menu.themes
    from pygame_menu.themes import Theme

    """
    Widgets: Menu widgets
    """
    import pygame_menu.widgets

"""
Version: Library version
"""
import pygame_menu.version

"""
Metadata: Information about the project
"""
__author__ = 'Pablo Pizarro R.'
__contributors__ = [

    # Author
    'ppizarror',

    # Contributors
    'anxuae',
    'apuly',
    'arpruss',
    'asierrayk',
    'DA820',
    'eforgacs',
    'i96751414',
    'ironsmile',
    'jwllee',
    'maditnerd',
    'MayuSakurai',
    'mrkprdo',
    'notrurs',
    'NullP01nt',
    'PandaRoux8',
    'Rifqi31',
    'ThePeeps191',
    'thisIsMikeKane',
    'vnmabus',
    'werdeil',
    'zPaw'

]
__copyright__ = 'Copyright 2017 Pablo Pizarro R. @ppizarror'
__description__ = 'A menu for pygame. Simple, and easy to use'
__email__ = 'pablo@ppizarror.com'
__keywords__ = 'pygame menu menus gui widget input button pygame-menu image sound ui'
__license__ = 'MIT'
__module_name__ = 'pygame-menu'
__url__ = 'https://pygame-menu.readthedocs.io'
__url_bug_tracker__ = 'https://github.com/ppizarror/pygame-menu/issues'
__url_documentation__ = 'https://pygame-menu.readthedocs.io'
__url_source_code__ = 'https://github.com/ppizarror/pygame-menu'
__version__ = pygame_menu.version.ver

"""
Print pygame-menu version.
"""
import os

if 'PYGAME_MENU_HIDE_VERSION' not in os.environ and 'PYGAME_HIDE_SUPPORT_PROMPT' not in os.environ:
    print(f'{__module_name__} {__version__}')

# Cleanup namespace
del os



"""
pygame-menu
https://github.com/ppizarror/pygame-menu

EXAMPLE - MAZE
Maze solver app, an improved version from https://github.com/ChrisKneller/pygame-pathfinder.
License: GNU General Public License v3.0
"""

__all__ = ['MazeApp']

import heapq
import pygame
import pygame_menu
import pygame_menu.utils as ut
import random
import time

from collections import deque
from math import inf
from typing import List, Union, Optional, Tuple, Any, Generator

from pygame_menu.examples import create_example_window

# Define some colors
BACKGROUND = (34, 40, 44)
BLACK = (0, 0, 0)
BLUE = (0, 0, 255)
BROWN = (186, 127, 50)
DARK_BLUE = (0, 0, 128)
DARK_GREEN = (0, 128, 0)
GREEN = (0, 255, 0)
GREY = (143, 143, 143)
LIGHT_BLUE = (0, 111, 255)
RED = (255, 0, 0)
WHITE = (255, 255, 255)


# noinspection PyMissingTypeHints,PyMissingOrEmptyDocstring
class AStarQueue(object):
    """
    A* Queue.
    """

    def __init__(self):
        self.myheap = []

    def show(self):
        return self.myheap

    def push(self, priority, distance, node):
        heapq.heappush(self.myheap, (priority, distance, node))

    def pop(self):
        priority, distance, node = heapq.heappop(self.myheap)
        return priority, distance, node


# noinspection PyMissingTypeHints,PyMissingOrEmptyDocstring
class PriorityQueue(object):
    """
    Priority Queue.
    """

    def __init__(self):
        self.myheap = []

    def show(self):
        return self.myheap

    def push(self, priority, node):
        heapq.heappush(self.myheap, (priority, node))

    def pop(self):
        priority, node = heapq.heappop(self.myheap)
        return priority, node


# noinspection PyMissingTypeHints,PyMissingOrEmptyDocstring
class PrioritySet(object):
    """
    Create a priority queue that doesn't add duplicate nodes.
    """

    def __init__(self):
        self.myheap = []
        self.myset = set()

    def show(self):
        return self.myheap

    def push(self, priority, node):
        if node not in self.myset:
            heapq.heappush(self.myheap, (priority, node))
            self.myset.add(node)

    def pop(self):
        priority, node = heapq.heappop(self.myheap)
        self.myset.remove(node)
        return priority, node


# noinspection PyDefaultArgument
class Node(object):
    """
    Make it easier to add different node types.
    """
    nodetypes = ['blank', 'start', 'end', 'wall', 'mud', 'dormant']

    colors = {
        'regular': {'blank': WHITE, 'start': RED, 'end': LIGHT_BLUE, 'wall': BLACK, 'mud': BROWN, 'dormant': GREY},
        'visited': {'blank': GREEN, 'start': RED, 'end': LIGHT_BLUE, 'wall': BLACK, 'mud': DARK_GREEN, 'dormant': GREY},
        'path': {'blank': BLUE, 'start': RED, 'end': LIGHT_BLUE, 'wall': BLACK, 'mud': DARK_BLUE, 'dormant': GREY}
    }

    distance_modifiers = {'blank': 1, 'start': 1, 'end': 1, 'wall': inf, 'mud': 3, 'dormant': inf}

    def __init__(self, nodetype: str, colors: dict = colors, dmf: dict = distance_modifiers) -> None:
        """
        Constructor.

        :param nodetype: Node type
        :param colors: Colors
        :param dmf: Distance modifier
        """
        self.nodetype = nodetype
        self.rcolor = colors['regular'][self.nodetype]
        self.vcolor = colors['visited'][self.nodetype]
        self.pcolor = colors['path'][self.nodetype]
        self.is_visited = True if nodetype == 'start' else True if nodetype == 'end' else False
        self.is_path = True if nodetype == 'start' else True if nodetype == 'end' else False
        self.distance_modifier = dmf[self.nodetype]
        self.color = self.pcolor if self.is_path else self.vcolor if self.is_visited else self.rcolor

    def update(
        self,
        nodetype: Union[bool, str] = False,
        is_visited: Union[bool, str] = 'unchanged',
        is_path: Union[bool, str] = 'unchanged',
        colors: dict = colors,
        dmf: dict = distance_modifiers,
        nodetypes: dict = nodetypes
    ) -> None:
        """
        Update the node.
        """
        if nodetype:
            assert nodetype in nodetypes, f"nodetype must be one of: {nodetypes}"
            if (self.nodetype == ('start' or 'end')) and (nodetype == ('wall' or 'mud')):
                pass
            else:
                self.nodetype = nodetype

        if is_visited != 'unchanged':
            assert type(is_visited) == bool, "'is_visited' must be boolean: True or False"
            self.is_visited = is_visited

        if is_path != 'unchanged':
            assert type(is_path) == bool, "'is_path' must be boolean: True or False"
            self.is_path = is_path

        self.rcolor = colors['regular'][self.nodetype]
        self.vcolor = colors['visited'][self.nodetype]
        self.pcolor = colors['path'][self.nodetype]
        self.distance_modifier = dmf[self.nodetype]
        self.color = self.pcolor if self.is_path else self.vcolor if self.is_visited else self.rcolor


_MazeType = List[List['Node']]
_Point2 = Tuple[int, int]


class MazeApp(object):
    """
    Maze class.
    """

    _algorithms: dict
    _diagonals: bool
    _end_point: _Point2
    _grid: _MazeType
    _height: int
    _margin: int
    _menu: 'pygame_menu.Menu'
    _mouse_drag: bool
    _offset: _Point2
    _rows: int
    _start_point: _Point2
    _visualize: bool
    _width: int

    def __init__(
        self,
        width: int = 8,
        rows: int = 75,
        margin: int = 0
    ) -> None:
        """
        Creates the maze.

        :param width: Width of each cell
        :param rows: Number of rows
        :param margin: Margin between cells
        """
        # Create the maze
        self._height = width  # so they are squares
        self._margin = margin
        self._offset = (25, 25)  # Maze offset within window
        self._width = width

        # Create a 2 dimensional array (a list of lists)
        self._grid = []
        self._rows = rows
        self._screen_width = self._rows * (self._width + self._margin) + self._margin * 2

        # Iterate through every row and column, adding blank nodes
        for row in range(rows):
            self._grid.append([])
            for column in range(rows):
                self._grid[row].append(Node('blank'))

        # Set start and end points for the pathfinder
        self._start_point = (random.randrange(2, rows - 1, 2) - 1, random.randrange(2, rows - 1, 2) - 1)
        self._end_point = (random.randrange(2, rows - 1, 2), random.randrange(2, rows - 1, 2))

        self._diagonals = True
        self._visualize = False

        # Used for handling click & drag
        self._mouse_drag = False
        self._drag_start_point = False
        self._drag_end_point = False

        # Used for deciding what to do in different situations
        self._algorithm_run = False
        self._algorithms = {0: 'dijkstra', 1: 'astar', 2: 'dfs', 3: 'bfs'}
        self._path_found = False

        # Create the window
        self._clock = pygame.time.Clock()
        self._fps = 60
        self._surface = create_example_window('Example - Maze', (900, 650))

        # Setups the menu
        self._setup_menu()

    def _setup_menu(self) -> None:
        """
        Setups the menu.
        """

        # Creates the events
        # noinspection PyUnusedLocal
        def onchange_dropselect(*args) -> None:
            """
            Called if the select is changed.
            """
            b = self._menu.get_widget('run_generator')
            b.readonly = False
            b.is_selectable = True
            b.set_cursor(pygame_menu.locals.CURSOR_HAND)

        def button_onmouseover(w: 'pygame_menu.widgets.Widget', _) -> None:
            """
            Set the background color of buttons if entered.
            """
            if w.readonly:
                return
            w.set_background_color((98, 103, 106))

        def button_onmouseleave(w: 'pygame_menu.widgets.Widget', _) -> None:
            """
            Set the background color of buttons if leaved.
            """
            w.set_background_color((75, 79, 81))

        def button_onmouseover_clear(w: 'pygame_menu.widgets.Widget', _) -> None:
            """
            Set the background color of buttons if entered.
            """
            if w.readonly:
                return
            w.set_background_color((139, 0, 0))

        def button_onmouseleave_clear(w: 'pygame_menu.widgets.Widget', _) -> None:
            """
            Set the background color of buttons if leaved.
            """
            w.set_background_color((205, 92, 92))

        def _visualize(value: bool) -> None:
            """
            Changes visualize.
            """
            self._visualize = value

        def _diagonals(value: bool) -> None:
            """
            Changes diagonals
            """
            self._diagonals = value
            if self._algorithm_run:
                self._path_found = self._update_path()

        theme = pygame_menu.Theme(
            background_color=pygame_menu.themes.TRANSPARENT_COLOR,
            title=False,
            widget_font=pygame_menu.font.FONT_FIRACODE,
            widget_font_color=(255, 255, 255),
            widget_margin=(0, 15),
            widget_selection_effect=pygame_menu.widgets.NoneSelection()
        )
        self._menu = pygame_menu.Menu(
            height=self._screen_width,
            mouse_motion_selection=True,
            position=(645, 25, False),
            theme=theme,
            title='',
            width=240
        )

        self._menu.add.toggle_switch(
            'Visualize',
            self._visualize,
            font_size=20,
            margin=(0, 5),
            onchange=_visualize,
            state_text_font_color=((0, 0, 0), (0, 0, 0)),
            state_text_font_size=15,
            switch_margin=(15, 0),
            width=80
        )
        self._menu.add.toggle_switch(
            'Diagonals',
            self._diagonals,
            font_size=20,
            margin=(0, 30),
            onchange=_diagonals,
            state_text_font_color=((0, 0, 0), (0, 0, 0)),
            switch_margin=(15, 0),
            toggleswitch_id='diagonals',
            width=80
        )

        self._menu.add.label(
            'Maze generator',
            font_name=pygame_menu.font.FONT_FIRACODE_BOLD,
            font_size=22,
            margin=(0, 5)
        ).translate(-12, 0)
        self._menu.add.dropselect(
            title='',
            items=[('Prim', 0),
                   ('Alt Prim', 1),
                   ('Recursive', 2),
                   ('(+) Terrain', 3)],
            dropselect_id='generator',
            font_size=16,
            onchange=onchange_dropselect,
            padding=0,
            placeholder='Select one',
            selection_box_height=5,
            selection_box_inflate=(0, 20),
            selection_box_margin=0,
            selection_box_text_margin=10,
            selection_box_width=200,
            selection_option_font_size=20,
            shadow_width=20
        )
        self._menu.add.vertical_margin(10)
        btn = self._menu.add.button(
            'Run Generator',
            self._run_generator,
            button_id='run_generator',
            font_size=20,
            margin=(0, 30),
            shadow_width=10,
        )
        btn.readonly = True
        btn.is_selectable = False

        self._menu.add.label(
            'Maze Solver',
            font_name=pygame_menu.font.FONT_FIRACODE_BOLD,
            font_size=22,
            margin=(0, 5)
        ).translate(-30, 0)
        self._menu.add.dropselect(
            title='',
            items=[('Dijkstra', 0),
                   ('A*', 1),
                   ('DFS', 2),
                   ('BFS', 3)],
            default=0,
            dropselect_id='solver',
            font_size=16,
            padding=0,
            placeholder='Select one',
            selection_box_height=5,
            selection_box_inflate=(0, 20),
            selection_box_margin=0,
            selection_box_text_margin=10,
            selection_box_width=200,
            selection_option_font_size=20,
            shadow_width=20
        )
        self._menu.add.vertical_margin(10)
        self._menu.add.button(
            'Run Solver',
            self._run_solver,
            button_id='run_solver',
            cursor=pygame_menu.locals.CURSOR_HAND,
            font_size=20,
            margin=(0, 75),
            shadow_width=10,
        )

        # Clears
        btn = self._menu.add.button(
            'Clear',
            self._clear_maze,
            background_color=(205, 92, 92),
            button_id='clear',
            cursor=pygame_menu.locals.CURSOR_HAND,
            font_size=20,
            margin=(0, 30),
            shadow_width=10,
        )
        btn.set_onmouseover(button_onmouseover_clear)
        btn.set_onmouseleave(button_onmouseleave_clear)
        btn.translate(-50, 0)

        # Create about menu
        menu_about = pygame_menu.Menu(
            height=self._screen_width + 20,
            mouse_motion_selection=True,
            position=(645, 8, False),
            theme=theme,
            title='',
            width=240
        )
        menu_about.add.label('pygame-menu\nMaze', font_name=pygame_menu.font.FONT_FIRACODE_BOLD, font_size=25,
                             margin=(0, 5))
        menu_about.add.vertical_margin(10)
        text = 'Left click to create a wall or move the start and end points.\n' \
               'Hold left CTRL and left click to create a sticky mud patch (whi' \
               'ch reduces movement speed to 1/3).\n'
        text += 'The point of these mud patches is to showcase Dijkstra\'s algor' \
                'ithm (first) and A* (second) by adjusting the "distances" betwe' \
                'en the nodes.\n\n'
        text += 'After a pathfinding algorithm has been run you can drag the sta' \
                'rt/end points around and see the visualisation update instantly' \
                ' for the new path using the algorithm that was last run.\n'
        menu_about.add.label(text, font_name=pygame_menu.font.FONT_FIRACODE, font_size=12,
                             margin=(0, 5), max_char=-1, padding=0)
        menu_about.add.label('License: GNU GPL v3.0', margin=(0, 5),
                             font_name=pygame_menu.font.FONT_FIRACODE, font_size=12)
        menu_about.add.url('https://github.com/ChrisKneller/pygame-pathfinder', 'ChrisKneller/pygame-pathfinder',
                           font_name=pygame_menu.font.FONT_FIRACODE, font_size=12,
                           font_color='#00bfff')
        menu_about.add.vertical_margin(20)
        menu_about.add.button(
            'Back',
            pygame_menu.events.BACK,
            button_id='about_back',
            cursor=pygame_menu.locals.CURSOR_HAND,
            font_size=20,
            shadow_width=10
        )

        btn = self._menu.add.button(
            'About',
            menu_about,
            button_id='about',
            float=True,
            font_size=20,
            margin=(0, 75),
            shadow_width=10
        )
        btn.translate(50, 0)

        # Configure buttons
        for btn in self._menu.get_widgets(['run_generator', 'run_solver', 'about', 'about_back']):
            btn.set_onmouseover(button_onmouseover)
            btn.set_onmouseleave(button_onmouseleave)
            if not btn.readonly:
                btn.set_cursor(pygame_menu.locals.CURSOR_HAND)
            btn.set_background_color((75, 79, 81))

    def _clear_maze(self) -> None:
        """
        Clear the maze.
        """
        self._path_found = False
        self._algorithm_run = False
        for row in range(self._rows):
            for column in range(self._rows):
                if (row, column) != self._start_point and (row, column) != self._end_point:
                    self._grid[row][column].update(nodetype='blank', is_visited=False, is_path=False)
        self._clear_visited()

    def _run_generator(self) -> None:
        """
        Run the generator.
        """
        if self._visualize:
            ut.set_pygame_cursor(pygame_menu.locals.CURSOR_NO)
        o_visualize = self._visualize
        gen_type = self._menu.get_widget('generator').get_value()[1]
        if gen_type != 3:
            self._clear_maze()
        if gen_type == 0:
            self._grid = self._prim(start_point=self._start_point)
        elif gen_type == 1:
            self._grid = self._better_prim(start_point=self._start_point)
        elif gen_type == 2:
            pygame.display.flip()
            self._recursive_division()
        elif gen_type == 3:
            self._random_terrain()
        self._visualize = o_visualize
        if self._visualize:
            ut.set_pygame_cursor(pygame_menu.locals.CURSOR_ARROW)

    def _run_solver(self) -> None:
        """
        Run the solver.
        """
        o_visualize = self._visualize
        solver_type = self._menu.get_widget('solver').get_value()[1]
        if self._visualize:
            pygame.display.flip()
        if self._path_found and self._algorithms[solver_type] == self._algorithm_run:
            self._visualize = False
        else:
            self._clear_visited()
            pygame.display.flip()
        if self._visualize:
            ut.set_pygame_cursor(pygame_menu.locals.CURSOR_NO)
        if solver_type == 0:
            self._path_found = self._dijkstra(self._grid, self._start_point, self._end_point)
            self._algorithm_run = 'dijkstra'
        elif solver_type == 1:
            self._path_found = self._dijkstra(self._grid, self._start_point, self._end_point, astar=True)
            self._algorithm_run = 'astar'
        elif solver_type == 2:
            self._path_found = self._xfs(self._grid, self._start_point, self._end_point, x='d')
            self._algorithm_run = 'dfs'
        elif solver_type == 3:
            self._path_found = self._xfs(self._grid, self._start_point, self._end_point, x='b')
            self._algorithm_run = 'bfs'
        self._visualize = o_visualize
        self._grid[self._start_point[0]][self._start_point[1]].update(nodetype='start')
        self._update_path()
        if self._visualize:
            ut.set_pygame_cursor(pygame_menu.locals.CURSOR_ARROW)

    def _check_esc(self) -> None:
        """
        Check if ESC button was pressed.
        """
        if self._visualize:
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    self._quit()
                if event.type == pygame.KEYDOWN and event.key == pygame.K_ESCAPE:
                    self._visualize = False

    @staticmethod
    def _sleep(ms: float) -> None:
        """
        Sleep time.

        :param ms: Sleep time in milliseconds
        """
        time.sleep(ms)

    def _better_prim(self, mazearray: Optional[_MazeType] = None, start_point: Optional[_Point2] = None) -> _MazeType:
        """
        Randomized Prim's algorithm for creating random mazes.
        This version maintains the traditional "maze" look, where a route cannot
        be diagonally connected to another point on the route.

        :param mazearray: Initial maze
        :param start_point: Creates a new starting point
        :return: New maze
        """
        # If a maze isn't input, we just create a grid full of walls
        if not mazearray:
            mazearray: _MazeType = []
            for row in range(self._rows):
                mazearray.append([])
                for column in range(self._rows):
                    if row % 2 != 0 and column % 2 != 0:
                        mazearray[row].append(Node('dormant'))
                    else:
                        mazearray[row].append(Node('wall'))
                    if self._visualize:
                        self._draw_square(mazearray, row, column)

        n = len(mazearray) - 1

        if not start_point:
            start_point = (random.randrange(1, n, 2), random.randrange(1, n, 2))
            mazearray[start_point[0]][start_point[1]].update(nodetype='blank')

        if self._visualize:
            self._draw_square(mazearray, start_point[0], start_point[1])
            pygame.display.flip()

        walls = set()

        starting_walls = self._get_neighbours(start_point, n)

        for wall, ntype in starting_walls:
            if mazearray[wall[0]][wall[1]].nodetype == 'wall':
                walls.add(wall)

        # While there are walls in the list (set):
        # Pick a random wall from the list. If only one of the cells that the wall divides is visited, then:
        # # Make the wall a passage and mark the unvisited cell as part of the maze.
        # # Add the neighboring walls of the cell to the wall list.
        # Remove the wall from the list.
        while len(walls) > 0:
            self._check_esc()
            wall = random.choice(tuple(walls))
            visited = 0
            add_to_maze = []

            for wall_neighbour, ntype in self._get_neighbours(wall, n):
                if mazearray[wall_neighbour[0]][wall_neighbour[1]].nodetype == 'blank':
                    visited += 1

            if visited <= 1:
                mazearray[wall[0]][wall[1]].update(nodetype='blank')

                if self._visualize:
                    self._draw_square(mazearray, wall[0], wall[1])
                    self._update_square(wall[0], wall[1])
                    self._sleep(0.0001)

                # A 'dormant' node (below) is a different type of node I had to create for this algo
                # otherwise the maze generated doesn't look like a traditional maze.
                # Every dormant eventually becomes a blank node, while the regular walls
                # sometimes become a passage between blanks and are sometimes left as walls
                for neighbour, ntype in self._get_neighbours(wall, n):
                    if mazearray[neighbour[0]][neighbour[1]].nodetype == 'dormant':
                        add_to_maze.append((neighbour[0], neighbour[1]))

                if len(add_to_maze) > 0:
                    cell = add_to_maze.pop()
                    mazearray[cell[0]][cell[1]].update(nodetype='blank')

                    if self._visualize:
                        self._draw_square(mazearray, cell[0], cell[1])
                        self._update_square(cell[0], cell[1])
                        self._sleep(0.0001)

                    for cell_neighbour, ntype in self._get_neighbours(cell, n):
                        if mazearray[cell_neighbour[0]][cell_neighbour[1]].nodetype == 'wall':
                            walls.add(cell_neighbour)

            walls.remove(wall)

        # Iterate through each dormant, if so, change to blank
        for row in range(self._rows):
            for column in range(self._rows):
                if mazearray[row][column].nodetype == 'dormant':
                    mazearray[row][column].update(nodetype='blank')

        mazearray[self._end_point[0]][self._end_point[1]].update(nodetype='end')
        mazearray[self._start_point[0]][self._start_point[1]].update(nodetype='start')

        return mazearray

    def _prim(self, mazearray: Optional[_MazeType] = None, start_point: Optional[_Point2] = None) -> _MazeType:
        """
        Randomized Prim's algorithm for creating random mazes.

        :param mazearray: Initial maze
        :param start_point: Creates a new starting point
        :return: New maze
        """
        # If a maze isn't input, we just create a grid full of walls
        if not mazearray:
            mazearray: _MazeType = []
            for row in range(self._rows):
                mazearray.append([])
                for column in range(self._rows):
                    mazearray[row].append(Node('wall'))
                    if self._visualize:
                        self._draw_square(mazearray, row, column)

        n = len(mazearray) - 1

        if not start_point:
            start_point = (random.randrange(0, n, 2), random.randrange(0, n, 2))

        if self._visualize:
            self._draw_square(mazearray, start_point[0], start_point[1])
            pygame.display.flip()

        walls = set([])

        neighbours = self._get_neighbours(start_point, n)

        for neighbour, ntype in neighbours:
            if mazearray[neighbour[0]][neighbour[1]].nodetype == 'wall':
                walls.add(neighbour)

        # While there are walls in the list:
        # Pick a random wall from the list. If only one of the cells that the wall divides is visited, then:
        # Make the wall a passage and mark the unvisited cell as part of the maze.
        # Add the neighboring walls of the cell to the wall list.
        # Remove the wall from the list.
        while len(walls) > 0:
            self._check_esc()
            wall = random.choice(tuple(walls))
            wall_neighbours = self._get_neighbours(wall, n)
            neighbouring_walls = set()
            pcount = 0
            for wall_neighbour, ntype in wall_neighbours:
                if wall_neighbour == (start_point or self._end_point):
                    continue
                if mazearray[wall_neighbour[0]][wall_neighbour[1]].nodetype != 'wall':
                    pcount += 1
                else:
                    neighbouring_walls.add(wall_neighbour)

            if pcount <= 1:
                mazearray[wall[0]][wall[1]].update(nodetype='blank')
                if self._visualize:
                    self._draw_square(mazearray, wall[0], wall[1])
                    self._update_square(wall[0], wall[1])
                    self._sleep(0.000001)

                walls.update(neighbouring_walls)

            walls.remove(wall)

        mazearray[self._end_point[0]][self._end_point[1]].update(nodetype='end')
        mazearray[self._start_point[0]][self._start_point[1]].update(nodetype='start')

        return mazearray

    def _recursive_division(
        self,
        chamber: Optional[Tuple[int, int, int, int]] = None,
        halving=True
    ) -> None:
        """
        Performs recursive division.

        :param chamber: Limit
        :param halving: Divide the recursion area by two
        """

        def _gaps_to_offset() -> List[int]:
            """
            Gaps to offset.
            """
            return [_x for _x in range(2, self._rows, 3)]

        gaps_to_offset = _gaps_to_offset()

        sleep = 0.001
        sleep_walls = 0.001

        # When no "chamber" is input,we are starting with the base grid
        if chamber is None:
            chamber_width = len(self._grid)
            chamber_height = len(self._grid[1])
            chamber_left = 0
            chamber_top = 0
        else:
            chamber_width = chamber[2]
            chamber_height = chamber[3]
            chamber_left = chamber[0]
            chamber_top = chamber[1]

        if halving:
            x_divide = int(chamber_width / 2)
            y_divide = int(chamber_height / 2)
        else:
            x_divide = chamber_width
            y_divide = chamber_height

        if chamber_width < 3:
            pass
        else:
            # draw x wall
            for y in range(chamber_height):
                self._grid[chamber_left + x_divide][chamber_top + y].update(nodetype='wall')
                self._draw_square(self._grid, chamber_left + x_divide, chamber_top + y)
                if self._visualize:
                    self._update_square(chamber_left + x_divide, chamber_top + y)
                    self._sleep(sleep_walls)

        if chamber_height < 3:
            pass
        else:
            # draw y wall
            for x in range(chamber_width):
                self._grid[chamber_left + x][chamber_top + y_divide].update(nodetype='wall')
                self._draw_square(self._grid, chamber_left + x, chamber_top + y_divide)
                if self._visualize:
                    self._update_square(chamber_left + x, chamber_top + y_divide)
                    self._sleep(sleep_walls)

        # Base case: stop dividing
        if chamber_width < 3 and chamber_height < 3:
            return

        # define the 4 new chambers (left, top, width, height)
        top_left = (chamber_left, chamber_top, x_divide, y_divide)
        top_right = (chamber_left + x_divide + 1, chamber_top, chamber_width - x_divide - 1, y_divide)
        bottom_left = (chamber_left, chamber_top + y_divide + 1, x_divide, chamber_height - y_divide - 1)
        bottom_right = (chamber_left + x_divide + 1, chamber_top + y_divide + 1, chamber_width - x_divide - 1,
                        chamber_height - y_divide - 1)

        chambers = (top_left, top_right, bottom_left, bottom_right)

        # define the 4 walls (of a + symbol) (left, top, width, height)
        left = (chamber_left, chamber_top + y_divide, x_divide, 1)
        right = (chamber_left + x_divide + 1, chamber_top + y_divide, chamber_width - x_divide - 1, 1)
        top = (chamber_left + x_divide, chamber_top, 1, y_divide)
        bottom = (chamber_left + x_divide, chamber_top + y_divide + 1, 1, chamber_height - y_divide - 1)

        walls = (left, right, top, bottom)

        gaps = 3
        for wall in random.sample(walls, gaps):
            if wall[3] == 1:
                x = random.randrange(wall[0], wall[0] + wall[2])
                y = wall[1]
                if x in gaps_to_offset and y in gaps_to_offset:
                    if wall[2] == x_divide:
                        x -= 1
                    else:
                        x += 1
                if x >= self._rows:
                    x = self._rows - 1
            else:  # the wall is horizontal
                x = wall[0]
                y = random.randrange(wall[1], wall[1] + wall[3])
                if y in gaps_to_offset and x in gaps_to_offset:
                    if wall[3] == y_divide:
                        y -= 1
                    else:
                        y += 1
                if y >= self._rows:
                    y = self._rows - 1
            self._grid[x][y].update(nodetype='blank')
            self._draw_square(self._grid, x, y)
            if self._visualize:
                self._update_square(x, y)
                self._sleep(sleep)

        # recursively apply the algorithm to all chambers
        for num, chamber in enumerate(chambers):
            self._recursive_division(chamber)
            self._check_esc()

    def _random_terrain(self, num_patches: Optional[int] = None) -> None:
        """
        Add random terrain to the maze.

        :param num_patches: Number of patches
        """
        if not num_patches:
            num_patches: int = random.randrange(int(self._rows / 10), int(self._rows / 4))

        terrain_nodes = set([])

        if self._visualize:
            pygame.display.flip()

        # For each patch we are creating we start with a centre node and branch outwards
        # getting neighbours of neighbours etc. for each node that we consider, there is
        # a variable probability of it becoming a patch of mud
        # As we branch outwards that probability decreases
        for patch in range(num_patches + 1):
            self._check_esc()
            neighbour_cycles = 0
            centre_point = (random.randrange(1, self._rows - 1), random.randrange(1, self._rows - 1))
            patch_type = 'mud'
            terrain_nodes.add(centre_point)

            while len(terrain_nodes) > 0:
                node = terrain_nodes.pop()

                if self._grid[node[0]][node[1]].nodetype != 'start' and self._grid[node[0]][node[1]].nodetype != 'end':
                    self._grid[node[0]][node[1]].update(nodetype=patch_type)
                    self._draw_square(self._grid, node[0], node[1])

                    if self._visualize:
                        self._update_square(node[0], node[1])
                        self._sleep(0.000001)

                neighbour_cycles += 1

                for node, ntype in self._get_neighbours(node):

                    if self._grid[node[0]][node[1]].nodetype == 'mud':
                        continue
                    threshold = 700 - (neighbour_cycles * 10)

                    if random.randrange(1, 101) <= threshold:
                        terrain_nodes.add(node)

    def _update_gui(self, draw_background=True, draw_menu=True, draw_grid=True) -> None:
        """
        Updates the gui.

        :param draw_background: Draw the background
        :param draw_menu: Draw the menu
        :param draw_grid: Draw the grid
        """
        if draw_background:
            # Draw a black background to set everything on
            self._surface.fill(BACKGROUND)

        if draw_grid:
            # Draw the grid
            for row in range(self._rows):
                for column in range(self._rows):
                    self._draw_square(self._grid, row, column)

        if draw_menu:
            self._menu.draw(self._surface)

    def _clear_visited(self) -> None:
        """
        Clear the visited grid.
        """
        excluded_nodetypes = ['start', 'end', 'wall', 'mud']
        for row in range(self._rows):
            for column in range(self._rows):
                if self._grid[row][column].nodetype not in excluded_nodetypes:
                    self._grid[row][column].update(nodetype='blank', is_visited=False, is_path=False)
                else:
                    self._grid[row][column].update(is_visited=False, is_path=False)
        self._update_gui()

    def _update_path(self) -> Union[bool, _MazeType]:
        """
        Updates the path.
        """
        self._clear_visited()
        assert self._algorithm_run in self._algorithms.values(), \
            f'last algorithm used ({self._algorithm_run}) is not in valid algorithms: {self._algorithms.values()}'

        visualize = self._visualize
        self._visualize = False
        if self._algorithm_run == 'dijkstra':
            path_found = self._dijkstra(self._grid, self._start_point, self._end_point)
        elif self._algorithm_run == 'astar':
            path_found = self._dijkstra(self._grid, self._start_point, self._end_point, astar=True)
        elif self._algorithm_run == 'dfs':
            path_found = self._xfs(self._grid, self._start_point, self._end_point, x='d')
        elif self._algorithm_run == 'bfs':
            path_found = self._xfs(self._grid, self._start_point, self._end_point, x='b')
        else:
            path_found = False
        self._visualize = visualize

        return path_found

    def _get_neighbours(
        self,
        node: _Point2,
        max_width: Optional[int] = None
    ) -> Generator[Tuple[_Point2, str], Any, None]:
        """
        Get the neighbours.

        :param node: Node
        :param max_width: Max width
        :return: Neighbours
        """
        if not max_width:
            max_width = self._rows - 1
        if not self._diagonals:
            neighbours = (
                ((min(max_width, node[0] + 1), node[1]), '+'),
                ((max(0, node[0] - 1), node[1]), '+'),
                ((node[0], min(max_width, node[1] + 1)), '+'),
                ((node[0], max(0, node[1] - 1)), '+')
            )
        else:
            neighbours = (
                ((min(max_width, node[0] + 1), node[1]), '+'),
                ((max(0, node[0] - 1), node[1]), '+'),
                ((node[0], min(max_width, node[1] + 1)), '+'),
                ((node[0], max(0, node[1] - 1)), '+'),
                ((min(max_width, node[0] + 1), min(max_width, node[1] + 1)), 'x'),
                ((min(max_width, node[0] + 1), max(0, node[1] - 1)), 'x'),
                ((max(0, node[0] - 1), min(max_width, node[1] + 1)), 'x'),
                ((max(0, node[0] - 1), max(0, node[1] - 1)), 'x')
            )

        # return neighbours
        return (neighbour for neighbour in neighbours if neighbour[0] != node)

    def _draw_square(self, grid: _MazeType, row: int, column: int) -> None:
        """
        Draws a square.

        :param grid: Grid
        :param row: Row number
        :param column: Column number
        """
        pygame.draw.rect(
            self._surface,
            grid[row][column].color,
            [
                (self._margin + self._width) * column + self._margin + self._offset[0],
                (self._margin + self._height) * row + self._margin + self._offset[1],
                self._width,
                self._height
            ]
        )
        pygame.event.pump()

    def _update_square(self, row: int, column: int) -> None:
        """
        Updates the square.

        :param row: Row number
        :param column: Column number
        :return:
        """
        # noinspection PyArgumentList
        pygame.display.update(
            (self._margin + self._width) * column + self._margin + self._offset[0],
            (self._margin + self._height) * row + self._margin + self._offset[1],
            self._width,
            self._height
        )
        pygame.event.pump()

    def _dijkstra(
        self,
        mazearray: _MazeType,
        start_point: _Point2 = (0, 0),
        goal_node: Optional[_Point2] = None,
        astar: bool = False
    ) -> bool:
        """
        Dijkstra's pathfinding algorithm, with the option to switch to A* by
        adding a heuristic of expected distance to end node.

        :param mazearray: Maze
        :param start_point: Starting point
        :param goal_node: Goal
        :param astar: A*
        :return: Finish status
        """
        heuristic = 0
        distance = 0

        # Get the dimensions of the (square) maze
        n = len(mazearray) - 1

        # Create the various data structures with speed in mind
        visited_nodes = set()
        unvisited_nodes = set([(x, y) for x in range(n + 1) for y in range(n + 1)])
        queue = AStarQueue()

        queue.push(distance + heuristic, distance, start_point)
        v_distances = {}

        # If a goal_node is not set, put it in the bottom right (1 square away from either edge)
        if not goal_node:
            goal_node = (n, n)
        priority, current_distance, current_node = queue.pop()
        start = time.perf_counter()

        # Main algorithm loop
        while current_node != goal_node and len(unvisited_nodes) > 0:
            self._check_esc()
            if current_node in visited_nodes:
                if len(queue.show()) == 0:
                    return False
                else:
                    priority, current_distance, current_node = queue.pop()
                    continue

            # Call to check neighbours of the current node
            for neighbour in self._get_neighbours(current_node, n):
                self._neighbours_loop(
                    neighbour,
                    mazearr=mazearray,
                    visited_nodes=visited_nodes,
                    unvisited_nodes=unvisited_nodes,
                    queue=queue,
                    # v_distances=v_distances,
                    # current_node=current_node,
                    current_distance=current_distance,
                    astar=astar
                )

            # When we have checked the current node, add and remove appropriately
            visited_nodes.add(current_node)
            unvisited_nodes.discard(current_node)

            # Add the distance to the visited distances' dictionary (used for traceback)
            v_distances[current_node] = current_distance

            # Pygame part: visited nodes mark visited nodes as green
            if (current_node[0], current_node[1]) != start_point:
                mazearray[current_node[0]][current_node[1]].update(is_visited=True)
                self._draw_square(mazearray, current_node[0], current_node[1])

                # If we want to visualise it (rather than run instantly)
                # then we update the grid with each loop
                if self._visualize:
                    self._update_square(current_node[0], current_node[1])
                    self._sleep(0.000001)

            # If there are no nodes in the queue then we return False (no path)
            if len(queue.show()) == 0:
                return False
            # Otherwise, we take the minimum distance as the new current node
            else:
                priority, current_distance, current_node = queue.pop()

        v_distances[goal_node] = current_distance + (1 if not self._diagonals else 2 ** 0.5)
        visited_nodes.add(goal_node)

        # Draw the path back from goal node to start node
        self._trace_back(goal_node, start_point, v_distances, n, mazearray)

        end = time.perf_counter()
        num_visited = len(visited_nodes)
        time_taken = end - start

        # Print timings
        print(f'Program finished in {time_taken:.4f}s after checking {num_visited}'
              f' nodes ({time_taken / num_visited:.8f} s/node)')

        return False if v_distances[goal_node] == float('inf') else True

    def _neighbours_loop(self, neighbour, mazearr, visited_nodes, unvisited_nodes, queue,  # v_distances, current_node,
                         current_distance, astar=False) -> None:
        """
        Loop through neighbours.
        """
        neighbour, ntype = neighbour
        heuristic = 0

        if astar:
            heuristic += abs(self._end_point[0] - neighbour[0]) + abs(self._end_point[1] - neighbour[1])
            heuristic *= 1  # If this goes above 1 then the shortest path is not guaranteed, but the attempted route becomes more direct

        # If the neighbour has already been visited
        if neighbour in visited_nodes:
            pass
        elif mazearr[neighbour[0]][neighbour[1]].nodetype == 'wall':
            visited_nodes.add(neighbour)
            unvisited_nodes.discard(neighbour)
        else:
            modifier = mazearr[neighbour[0]][neighbour[1]].distance_modifier
            if ntype == '+':
                queue.push(current_distance + (1 * modifier) + heuristic, current_distance + (1 * modifier), neighbour)
            elif ntype == 'x':
                queue.push(current_distance + ((2 ** 0.5) * modifier) + heuristic,
                           current_distance + ((2 ** 0.5) * modifier), neighbour)

    def _trace_back(self, goal_node, start_node, v_distances, n, mazearray) -> None:
        """
        (DIJKSTRA/A*) trace a path back from the end node to the start node after the algorithm has been run.
        """
        # Begin the list of nodes which will represent the path back, starting with the end node
        path = [goal_node]
        current_node = goal_node

        # Set the loop in motion until we get back to the start
        while current_node != start_node:
            # Start an empty priority queue for the current node to check all neighbours
            neighbour_distances = PriorityQueue()

            neighbours = self._get_neighbours(current_node, n)

            # Had some errors during testing, not sure if this is still necessary
            try:
                v_distances[current_node]
            except Exception as e:
                print(e)

            # For each neighbour of the current node, add its location and distance
            # to a priority queue
            for neighbour, ntype in neighbours:
                if neighbour in v_distances:
                    distance = v_distances[neighbour]
                    neighbour_distances.push(distance, neighbour)

            # Pop the lowest value off; that is the next node in our path
            distance, smallest_neighbour = neighbour_distances.pop()
            mazearray[smallest_neighbour[0]][smallest_neighbour[1]].update(is_path=True)

            # Update pygame display
            self._draw_square(mazearray, smallest_neighbour[0], smallest_neighbour[1])

            path.append(smallest_neighbour)
            current_node = smallest_neighbour

        pygame.display.flip()

        mazearray[start_node[0]][start_node[1]].update(is_path=True)

    def _xfs(self, mazearray: _MazeType, start_point: _Point2, goal_node: _Point2, x: str) -> bool:
        """
        This is a function where you choose x='b' or x='d' to run bfs (breadth-first search) or
        dfs (depth-first search) on your chosen mazearray (grid format), with chosen start_point (x,y)
        and chosen goal_node (x,y).

        :param mazearray: Maze array
        :param start_point: Starting point
        :param goal_node: Ending point
        :param x: Type
        :return: False
        """
        assert x == 'b' or x == 'd', "x should equal 'b' or 'd' to make this bfs or dfs"

        # Get the dimensions of the (square) maze
        n = len(mazearray) - 1

        # Create the various data structures with speed in mind
        mydeque = deque()
        mydeque.append(start_point)
        visited_nodes = set([])
        path_dict = {start_point: None}

        # Main algorithm loop
        while len(mydeque) > 0:
            self._check_esc()
            if x == 'd':
                current_node = mydeque.pop()
            elif x == 'b':
                current_node = mydeque.popleft()

            # noinspection PyUnboundLocalVariable
            if current_node == goal_node:
                # Trace back to start using path_dict
                path_node = goal_node
                while True:
                    path_node = path_dict[path_node]
                    mazearray[path_node[0]][path_node[1]].update(is_path=True)
                    self._draw_square(mazearray, path_node[0], path_node[1])
                    if self._visualize:
                        self._update_square(path_node[0], path_node[1])
                    if path_node == start_point:
                        return True

            if mazearray[current_node[0]][current_node[1]].nodetype == 'wall':
                continue

            if current_node not in visited_nodes:
                visited_nodes.add(current_node)
                mazearray[current_node[0]][current_node[1]].update(is_visited=True)
                self._draw_square(mazearray, current_node[0], current_node[1])
                if self._visualize:
                    self._update_square(current_node[0], current_node[1])
                    self._sleep(0.001)

                for neighbour, ntype in self._get_neighbours(current_node, n):
                    mydeque.append(neighbour)
                    # Used for tracing back
                    if neighbour not in visited_nodes:
                        path_dict[neighbour] = current_node

        pygame.display.flip()
        return False

    def _get_pos(self, pos: _Point2) -> _Point2:
        """
        Return the column/row position within maze.

        :param pos: Position
        :return: Position in col/row
        """
        column = (pos[0] - self._offset[0]) // (self._width + self._margin)
        row = (pos[1] - self._offset[1]) // (self._height + self._margin)
        return column, row

    def _pos_in_grid(self, pos: _Point2) -> bool:
        """
        Check if the mouse position is in grid.

        :param pos: Position
        :return: True if inside
        """
        x = pos[0] - self._offset[0]
        y = pos[1] - self._offset[1]
        return 1 <= x <= self._screen_width and 1 <= y <= self._screen_width

    @staticmethod
    def _quit() -> None:
        """
        Quit app.
        """
        pygame.quit()
        exit()

    def mainloop(self, test: bool) -> None:
        """
        Executes the main loop of the app.

        :param test: If True, runs only 1 frame
        """
        print('Press [ESC] to skip process if Visualize is On')
        while True:

            # Application events
            events = pygame.event.get()

            # Update the menu
            self._menu.update(events)

            # If a menu widget disable its active state, disable the events, this is due to
            # user can click outside a dropselection box, and that triggers to disable active
            # state. If so, the event is destroyed, thus avoiding clicking the canvas
            if pygame_menu.events.MENU_LAST_WIDGET_DISABLE_ACTIVE_STATE in self._menu.get_last_update_mode()[0]:
                events = []

            for event in events:
                # User closes
                if event.type == pygame.QUIT:
                    self._quit()

                # Write in the board
                elif event.type == pygame.MOUSEBUTTONDOWN:
                    pos = pygame.mouse.get_pos()

                    # Find out which keys have been pressed
                    pressed = pygame.key.get_pressed()

                    # If click is inside grid
                    if self._pos_in_grid(pos):

                        # Change the x/y screen coordinates to grid coordinates
                        column, row = self._get_pos(pos)

                        if (row, column) == self._start_point:
                            self._drag_start_point = True
                        elif (row, column) == self._end_point:
                            self._drag_end_point = True
                        else:
                            cell_updated = self._grid[row][column]
                            if pressed[pygame.K_LCTRL]:
                                update_cell_to = 'mud'
                            else:
                                update_cell_to = 'wall'
                            cell_updated.update(nodetype=update_cell_to)
                            self._mouse_drag = True
                            if self._algorithm_run and cell_updated.is_path:
                                self._path_found = self._update_path()

                # Turn off all mouse drags if mouse Button released
                elif event.type == pygame.MOUSEBUTTONUP:
                    self._mouse_drag = self._drag_end_point = self._drag_start_point = False

                # Moves the mouse
                elif event.type == pygame.MOUSEMOTION:
                    # Boolean values saying whether left, middle and right mouse buttons are currently pressed
                    left, middle, right = pygame.mouse.get_pressed()

                    # Sometimes we get stuck in this loop if the mousebutton is released while not in the pygame screen
                    # This acts to break out of that loop
                    if not left:
                        self._mouse_drag = self._drag_end_point = self._drag_start_point = False
                        continue

                    # User moves the mouse. Get the position
                    pos = pygame.mouse.get_pos()

                    # Change the x/y screen coordinates to grid coordinates
                    column, row = self._get_pos(pos)

                    # Turn mouse_drag off if mouse goes outside of grid
                    if not self._pos_in_grid(pos):
                        self._mouse_drag = False
                        continue

                    try:
                        cell_updated = self._grid[row][column]
                    except IndexError:
                        continue

                    # Add walls or sticky mud patches
                    pressed = pygame.key.get_pressed()
                    if self._mouse_drag:
                        if (row, column) == self._start_point:
                            pass
                        elif (row, column) == self._end_point:
                            pass
                        else:
                            if pressed[pygame.K_LCTRL]:
                                update_cell_to = 'mud'
                            else:
                                update_cell_to = 'wall'
                            cell_updated.update(nodetype=update_cell_to)

                        self._mouse_drag = True
                        if self._algorithm_run:
                            if cell_updated.is_path:
                                self._path_found = self._update_path()

                    # Move the start point
                    elif self._drag_start_point:
                        if self._grid[row][column].nodetype == 'blank':
                            self._grid[self._start_point[0]][self._start_point[1]].update(nodetype='blank',
                                                                                          is_path=False,
                                                                                          is_visited=False)
                            self._start_point = (row, column)
                            self._grid[self._start_point[0]][self._start_point[1]].update(nodetype='start')
                            # If we have already run the algorithm, update it as the point is moved
                            if self._algorithm_run:
                                self._path_found = self._update_path()
                                self._grid[self._start_point[0]][self._start_point[1]].update(nodetype='start')

                    # Move the end point
                    elif self._drag_end_point:
                        if self._grid[row][column].nodetype == 'blank':
                            self._grid[self._end_point[0]][self._end_point[1]].update(nodetype='blank', is_path=False,
                                                                                      is_visited=False)
                            self._end_point = (row, column)
                            self._grid[self._end_point[0]][self._end_point[1]].update(nodetype='end')
                            # If we have already run the algorithm, update it as the point is moved
                            if self._algorithm_run:
                                self._path_found = self._update_path()
                                self._grid[self._end_point[0]][self._end_point[1]].update(nodetype='start')

                    pygame.display.flip()

            # Update the app
            self._update_gui()

            self._grid[self._start_point[0]][self._start_point[1]].update(nodetype='start')
            self._grid[self._end_point[0]][self._end_point[1]].update(nodetype='end')

            # Flip surface
            pygame.display.flip()

            # Update clock
            self._clock.tick(self._fps)

            # At first loop returns
            if test:
                break


def main(test: bool = False) -> 'MazeApp':
    """
    Main function.

    :param test: Indicate function is being tested
    :return: App
    """
    app = MazeApp()
    app.mainloop(test)
    return app


if __name__ == '__main__':
    main()



"""
pygame-menu-ce
https://github.com/ppizarror/pygame-menu

SETUP DISTRIBUTION
Create setup for PyPi.

License:
-------------------------------------------------------------------------------
The MIT License (MIT)
Copyright 2017 Pablo Pizarro R. @ppizarror

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the Software
is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
-------------------------------------------------------------------------------
"""

from setuptools import setup, find_packages
import pygame_menu

# Load readme
with open('README.rst') as f:
    long_description = f.read()

# Load requirements
with open('requirements.txt') as f:
    requirements = []
    for line in f:
        requirements.append(line.strip())

# Setup library
setup(
    name=pygame_menu.__module_name__ + '-ce',
    version=pygame_menu.__version__,
    author=pygame_menu.__author__,
    author_email=pygame_menu.__email__,
    description=pygame_menu.__description__.replace('pygame', 'pygame-ce'),
    long_description=long_description,
    url=pygame_menu.__url__,
    project_urls={
        'Bug Tracker': pygame_menu.__url_bug_tracker__,
        'Documentation': pygame_menu.__url_documentation__,
        'Source Code': pygame_menu.__url_source_code__
    },
    license=pygame_menu.__license__,
    platforms=['any'],
    keywords=pygame_menu.__keywords__,
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python',
        'Topic :: Games/Entertainment',
        'Topic :: Multimedia',
        'Topic :: Software Development :: Libraries :: pygame',
        'Topic :: Text Processing'
    ],
    include_package_data=True,
    packages=find_packages(exclude=['test']),
    python_requires='>=3.7, <4',
    install_requires=requirements,
    extras_require={
        'docs': ['sphinx', 'sphinx-autodoc-typehints>=1.2.0', 'sphinx-rtd-theme'],
        'test': ['nose2[coverage_plugin]']
    },
    setup_requires=[
        'setuptools',
    ],
    options={
        'bdist_wheel': {'universal': False}
    },
    entry_points={
        'pyinstaller40': ['hook-dirs = pygame_menu.__pyinstaller:get_hook_dirs']
    }
)



"""
pygame-menu
https://github.com/ppizarror/pygame-menu

EXAMPLE - TIMER CLOCK
Example file, timer clock with in-menu options.
"""

__all__ = ['main']

import pygame
import pygame_menu
from pygame_menu.examples import create_example_window

import datetime
from random import randrange
from typing import List, Tuple, Optional

# Constants and global variables
ABOUT = [f'pygame-menu {pygame_menu.__version__}',
         f'Author: {pygame_menu.__author__}',
         '',
         f'Email: {pygame_menu.__email__}']
COLOR_BACKGROUND = [128, 0, 128]
FPS = 60
H_SIZE = 600  # Height of window size
HELP = ['Press ESC to enable/disable Menu',
        'Press ENTER to access a Sub-Menu or use an option',
        'Press UP/DOWN to move through Menu',
        'Press LEFT/RIGHT to move through Selectors']
W_SIZE = 800  # Width of window size

surface: Optional['pygame.Surface'] = None
timer: Optional[List[float]] = None


def mainmenu_background() -> None:
    """
    Background color of the main menu, on this function user can plot
    images, play sounds, etc.
    """
    surface.fill((40, 0, 40))


def reset_timer() -> None:
    """
    Reset timer.
    """
    timer[0] = 0


class TestCallClassMethod(object):
    """
    Class call method.
    """

    @staticmethod
    def update_game_settings() -> None:
        """
        Class method.
        """
        print('Update game with new settings')


def change_color_bg(value: Tuple, c: Optional[Tuple] = None, **kwargs) -> None:
    """
    Change background color.

    :param value: Selected option (data, index)
    :param c: Color tuple
    """
    color, _ = value
    if c == (-1, -1, -1):  # If random color
        c = (randrange(0, 255), randrange(0, 255), randrange(0, 255))
    if kwargs['write_on_console']:
        print('New background color: {0} ({1},{2},{3})'.format(color[0], *c))
    COLOR_BACKGROUND[0] = c[0]
    COLOR_BACKGROUND[1] = c[1]
    COLOR_BACKGROUND[2] = c[2]


def main(test: bool = False) -> None:
    """
    Main program.

    :param test: Indicate function is being tested
    """

    # -------------------------------------------------------------------------
    # Init
    # -------------------------------------------------------------------------
    # Write help message on console
    for m in HELP:
        print(m)

    # Create window
    global surface
    surface = create_example_window('Example - Timer Clock', (W_SIZE, H_SIZE))

    # Main timer and game clock
    clock = pygame.time.Clock()
    global timer
    timer = [0]
    dt = 1.0 / FPS
    timer_font = pygame_menu.font.get_font(pygame_menu.font.FONT_NEVIS, 100)
    frame = 0

    # -------------------------------------------------------------------------
    # Create menus: Timer
    # -------------------------------------------------------------------------

    timer_theme = pygame_menu.themes.THEME_DARK.copy()  # Create a new copy
    timer_theme.background_color = (0, 0, 0, 180)  # Enable transparency

    # Timer
    timer_menu = pygame_menu.Menu(
        height=400,
        onclose=pygame_menu.events.RESET,
        theme=timer_theme,
        title='Timer Menu',
        width=600
    )

    # Add widgets
    timer_menu.add.button('Reset timer', reset_timer)

    # Adds a selector (element that can handle functions)
    timer_menu.add.selector(
        title='Change color ',
        items=[('Random', (-1, -1, -1)),  # Values of selector, call to change_color_bg
               ('Default', (128, 0, 128)),
               ('Black', (0, 0, 0)),
               ('Blue', (12, 12, 200))],
        default=1,  # Optional parameter that sets default item of selector
        onchange=change_color_bg,  # Action when changing element with left/right
        onreturn=change_color_bg,  # Action when pressing return on an element
        # All the following kwargs are passed to change_color_bg function
        write_on_console=True
    )
    timer_menu.add.button('Update game object', TestCallClassMethod().update_game_settings)
    timer_menu.add.button('Return to Menu', pygame_menu.events.BACK)
    timer_menu.add.button('Close Menu', pygame_menu.events.CLOSE)

    # -------------------------------------------------------------------------
    # Create menus: Help
    # -------------------------------------------------------------------------
    help_theme = pygame_menu.Theme(
        background_color=(30, 50, 107, 190),  # 75% opacity
        title_background_color=(120, 45, 30, 190),
        title_font=pygame_menu.font.FONT_FRANCHISE,
        title_font_size=60,
        widget_font=pygame_menu.font.FONT_FRANCHISE,
        widget_font_color=(170, 170, 170),
        widget_font_shadow=True,
        widget_font_shadow_position=pygame_menu.locals.POSITION_SOUTHEAST,
        widget_font_size=45
    )

    help_menu = pygame_menu.Menu(
        height=600,  # Fullscreen
        theme=help_theme,
        title='Help',
        width=800
    )
    for m in HELP:
        help_menu.add.label(m, align=pygame_menu.locals.ALIGN_CENTER)
    help_menu.add.vertical_margin(25)
    help_menu.add.button('Return to Menu', pygame_menu.events.BACK)

    # -------------------------------------------------------------------------
    # Create menus: About
    # -------------------------------------------------------------------------
    about_theme = pygame_menu.themes.THEME_DARK.copy()
    about_theme.widget_font = pygame_menu.font.FONT_NEVIS
    about_theme.title_font = pygame_menu.font.FONT_8BIT
    about_theme.title_offset = (5, -2)
    about_theme.widget_offset = (0, 0.14)

    about_menu = pygame_menu.Menu(
        center_content=False,
        height=400,
        mouse_visible=False,
        theme=about_theme,
        title='About',
        width=600
    )
    for m in ABOUT:
        about_menu.add.label(m, margin=(0, 0))
    about_menu.add.label('')
    about_menu.add.button('Return to Menu', pygame_menu.events.BACK)

    # -------------------------------------------------------------------------
    # Create menus: Main menu
    # -------------------------------------------------------------------------
    main_menu = pygame_menu.Menu(
        enabled=False,
        height=400,
        theme=pygame_menu.themes.THEME_DARK,
        title='Main Menu',
        width=600
    )

    main_menu.add.button(timer_menu.get_title(), timer_menu)  # Add timer submenu
    main_menu.add.button(help_menu.get_title(), help_menu)  # Add help submenu
    main_menu.add.button(about_menu.get_title(), about_menu)  # Add about submenu
    main_menu.add.button('Exit', pygame_menu.events.EXIT)  # Add exit function

    # -------------------------------------------------------------------------
    # Main loop
    # -------------------------------------------------------------------------
    while True:

        # Tick clock
        clock.tick(FPS)
        timer[0] += dt
        frame += 1

        # Title is evaluated at current level as the title of the base pointer
        # object (main_menu) can change if user opens submenus
        current_menu = main_menu.get_current()
        if current_menu.get_title() != 'Main Menu' or not main_menu.is_enabled():
            # Draw timer
            surface.fill(COLOR_BACKGROUND)
            time_string = str(datetime.timedelta(seconds=int(timer[0])))
            time_blit = timer_font.render(time_string, True, (255, 255, 255))
            time_blit_size = time_blit.get_size()
            surface.blit(time_blit, (int(W_SIZE / 2 - time_blit_size[0] / 2), int(H_SIZE / 2 - time_blit_size[1] / 2)))
        else:
            # Background color if the menu is enabled and timer is hidden
            surface.fill((40, 0, 40))

        # Application events
        events = pygame.event.get()
        for event in events:
            if event.type == pygame.QUIT:
                exit()
            elif event.type == pygame.KEYDOWN:
                if event.key == pygame.K_ESCAPE and current_menu.get_title() == 'Main Menu':
                    main_menu.toggle()

        if main_menu.is_enabled():
            main_menu.draw(surface)
            main_menu.update(events)

        # Flip surface
        pygame.display.flip()

        # At first loop returns
        if test and frame == 2:
            break


if __name__ == '__main__':
    main()



"""
pygame-menu
https://github.com/ppizarror/pygame-menu

EXAMPLE - WINDOW RESIZE
Resize the menu when the window is resized.
"""

import pygame
import pygame_menu

pygame.init()

surface = pygame.display.set_mode((600, 400), pygame.RESIZABLE)
pygame.display.set_caption("Example resizable window")

menu = pygame_menu.Menu(
    height=100,
    theme=pygame_menu.themes.THEME_BLUE,
    title='Welcome',
    width=100
)


def on_resize() -> None:
    """
    Function checked if the window is resized.
    """
    window_size = surface.get_size()
    new_w, new_h = 0.75 * window_size[0], 0.7 * window_size[1]
    menu.resize(new_w, new_h)
    print(f'New menu size: {menu.get_size()}')


menu.add.label('Resize the window!')
user_name = menu.add.text_input('Name: ', default='John Doe', maxchar=10)
menu.add.selector('Difficulty: ', [('Hard', 1), ('Easy', 2)])
menu.add.button('Quit', pygame_menu.events.EXIT)
menu.enable()
on_resize()  # Set initial size

if __name__ == '__main__':
    while True:
        events = pygame.event.get()
        for event in events:
            if event.type == pygame.QUIT:
                pygame.quit()
                break
            if event.type == pygame.VIDEORESIZE:
                # Update the surface
                surface = pygame.display.set_mode((event.w, event.h), pygame.RESIZABLE)
                # Call the menu event
                on_resize()

        # Draw the menu
        surface.fill((25, 0, 50))

        menu.update(events)
        menu.draw(surface)

        pygame.display.flip()



"""
pygame-menu
https://github.com/ppizarror/pygame-menu

VERTICAL VILL
Vertical fill box. Fills all available vertical space.
"""

__all__ = [
    'VFill',
    'VFillManager'
]

import pygame
import pygame_menu

from abc import ABC
from pygame_menu.widgets.core.widget import AbstractWidgetManager
from pygame_menu.widgets.widget.none import NoneWidget

from pygame_menu._types import NumberType, NumberInstance


# noinspection PyMissingOrEmptyDocstring
class VFill(NoneWidget):
    """
    Vertical fill widget. This widget fills all vertical space if available, else,
    it uses the min height.

    .. note::

        VFill does not accept any transformation.

    :param min_height: Minimum height in px
    :param widget_id: ID of the widget
    """
    _min_height: int

    def __init__(
        self,
        min_height: NumberType,
        widget_id: str = ''
    ) -> None:
        assert isinstance(min_height, NumberInstance)
        assert min_height >= 0, 'negative min height is not valid'
        super(VFill, self).__init__(widget_id=widget_id)
        self._min_height = min_height

    def get_rect(self, *args, **kwargs) -> 'pygame.Rect':
        # Get all menu widgets, and for those in the same column store the total
        # size (without considering other vfills). Then, divide all available height
        # in the total vfills found
        c = self.get_col_row_index()[0]
        if c == -1:
            self._rect.height = self._min_height
            return self._rect
        menu_widgets_col = self.get_menu().get_widgets_column(c)
        available_height = self.get_menu().get_height(inner=True)
        total_column_height = 0
        total_vfills = 0
        column_vfills = []

        for w in menu_widgets_col:
            if not isinstance(w, VFill):
                total_column_height += w.get_height()
            else:
                column_vfills.append(w)
                total_column_height += w._min_height
                total_vfills += 1

        vfill_height = ((available_height - total_column_height) / total_vfills) if total_vfills > 0 else 0

        # Discount 1px only to last
        if self == column_vfills[-1] and vfill_height > 0:
            vfill_height -= 1

        vfill_height = max(0, vfill_height)
        self._rect.height = self._min_height + vfill_height

        return self._rect


class VFillManager(AbstractWidgetManager, ABC):
    """
    VFill manager.
    """

    def vertical_fill(
        self,
        min_height: NumberType = 0,
        vfill_id: str = ''
    ) -> 'pygame_menu.widgets.VFill':
        """
        Adds a vertical fill to the Menu. This widget fills all vertical space
        if available, else, it uses the min height.

        .. note::

            This is applied only to the base Menu (not the currently displayed,
            stored in ``_current`` pointer); for such behaviour apply to
            :py:meth:`pygame_menu.menu.Menu.get_current` object.

        :param min_height: Minimum height in px
        :param vfill_id: ID of the vertical fill
        :return: Widget object
        :rtype: :py:class:`pygame_menu.widgets.VFill`
        """
        attributes = self._filter_widget_attributes({})
        widget = VFill(min_height, widget_id=vfill_id)
        self._configure_widget(widget=widget, **attributes)
        self._append_widget(widget)
        return widget



"""
pygame-menu
https://github.com/ppizarror/pygame-menu

PROGRESS BAR
Progress bar widget.
"""

__all__ = [

    # Class
    'ProgressBar',
    'ProgressBarManager',

    # Types
    'ProgressBarTextFormatType'

]

import pygame
import pygame_menu

from abc import ABC
from pygame_menu.font import FontType, assert_font
from pygame_menu.locals import ALIGN_LEFT, ALIGN_CENTER
from pygame_menu.utils import assert_color, assert_vector, make_surface, \
    assert_alignment, parse_padding
from pygame_menu.widgets.core.widget import Widget, WidgetTransformationNotImplemented, \
    AbstractWidgetManager

from pygame_menu._types import Any, CallbackType, Optional, ColorType, NumberType, \
    Tuple2IntType, NumberInstance, ColorInputType, EventVectorType, Callable, \
    PaddingType, Tuple4IntType

ProgressBarTextFormatType = Callable[[NumberType], str]


# noinspection PyMissingOrEmptyDocstring
class ProgressBar(Widget):
    """
    Progress bar widget, offers a bar that accepts a percentage from ``0`` to ``100``.

    .. note::

        ProgressBar only accepts translation transformation.

    :param title: Progressbar title
    :param progressbar_id: ProgressBar ID
    :param default: Default value of the progressbar, from ``0`` to ``100``
    :param width: Progress bar width in px
    :param onselect: Function when selecting the widget
    :param box_background_color: Background color of the box
    :param box_border_color: Border color of the box
    :param box_border_width: Border width of the box in px
    :param box_margin: Box margin on x-axis and y-axis (x, y) respect to the title of the widget in px
    :param box_progress_color: Box progress color
    :param box_progress_padding: Box progress padding
    :param progress_text_align: Align of the progress text, can be CENTER, LEFT or RIGHT. See :py:mod:`pygame_menu.locals`
    :param progress_text_enabled: Enables the progress text over box
    :param progress_text_font: Progress font. If ``None`` uses the same as the widget font
    :param progress_text_font_color: Progress font color. If ``None`` uses the same as the widget font
    :param progress_text_font_hfactor: Height factor of the font height relative to the widget font height
    :param progress_text_format: Format function of the progress text, which considers as input the progress value (0-100)
    :param progress_text_margin: Margin of the progress box on x-axis and y-axis in px
    :param progress_text_placeholder: Placeholder of the progress text, which considers as format the output of ``progress_text_format``
    :param args: Optional arguments for callbacks
    :param kwargs: Optional keyword arguments
    """
    _box: 'pygame.Surface'
    _box_background_color: ColorType
    _box_border_color: ColorType
    _box_border_width: int
    _box_height: int
    _box_margin: Tuple2IntType
    _box_pos: int
    _box_progress_color: ColorType
    _box_progress_padding: Tuple4IntType
    _progress: NumberType
    _progress_font: FontType
    _progress_text_align: str
    _progress_text_enabled: bool
    _progress_text_font: Optional[FontType]
    _progress_text_font_color: ColorType
    _progress_text_font_height: int
    _progress_text_font_height_factor: float
    _progress_text_format: ProgressBarTextFormatType
    _progress_text_margin: Tuple2IntType
    _progress_text_placeholder: str
    _width: int

    def __init__(
        self,
        title: Any,
        progressbar_id: str = '',
        default: NumberType = 0,
        width: int = 150,
        onselect: CallbackType = None,
        box_background_color: ColorInputType = (255, 255, 255),
        box_border_color: ColorInputType = (0, 0, 0),
        box_border_width: int = 1,
        box_margin: Tuple2IntType = (25, 0),
        box_progress_color: ColorInputType = (0, 255, 0),
        box_progress_padding: PaddingType = (1, 1),
        progress_text_align: str = ALIGN_CENTER,
        progress_text_enabled: bool = True,
        progress_text_font: Optional[FontType] = None,
        progress_text_font_color: ColorInputType = (0, 0, 0),
        progress_text_font_hfactor: float = 0.8,
        progress_text_format: ProgressBarTextFormatType = lambda x: str(round(x, 1)),
        progress_text_margin: Tuple2IntType = (0, 0),
        progress_text_placeholder: str = '{0} %',
        *args,
        **kwargs
    ) -> None:
        super(ProgressBar, self).__init__(
            args=args,
            kwargs=kwargs,
            onselect=onselect,
            title=title,
            widget_id=progressbar_id
        )

        # Check the value
        assert isinstance(default, NumberInstance)
        assert 0 <= default <= 100, 'default value must range from 0 to 100'

        # Check fonts
        if progress_text_font is not None:
            assert_font(progress_text_font)
        assert isinstance(progress_text_font_hfactor, NumberInstance)
        assert progress_text_font_hfactor > 0, \
            'progress text font height factor must be greater than zero'

        # Check colors
        box_background_color = assert_color(box_background_color)
        box_border_color = assert_color(box_border_color)
        box_progress_color = assert_color(box_progress_color)
        progress_text_font_color = assert_color(progress_text_font_color)

        # Check dimensions and sizes
        assert isinstance(box_border_width, int)
        assert box_border_width >= 0, \
            'box border width must be equal or greater than zero'
        assert_vector(box_margin, 2, int)
        assert_vector(progress_text_margin, 2, int)
        assert isinstance(width, int)
        assert width > 0, \
            'width must be greater than zero'
        box_progress_padding = parse_padding(box_progress_padding)
        self._box_progress_padding = box_progress_padding

        # Check progress text
        assert isinstance(progress_text_enabled, bool)
        assert callable(progress_text_format)
        assert isinstance(progress_text_format(0), str)
        assert isinstance(progress_text_placeholder, str)
        assert_alignment(progress_text_align)

        # Store properties
        self._default_value = default
        self._box_background_color = box_background_color
        self._box_border_color = box_border_color
        self._box_border_width = box_border_width
        self._box_margin = box_margin
        self._box_progress_color = box_progress_color
        self._progress = default
        self._progress_text_align = progress_text_align
        self._progress_text_enabled = progress_text_enabled
        self._progress_text_font = progress_text_font
        self._progress_text_font_color = progress_text_font_color
        self._progress_text_font_height = 0
        self._progress_text_font_height_factor = progress_text_font_hfactor
        self._progress_text_format = progress_text_format
        self._progress_text_margin = progress_text_margin
        self._progress_text_placeholder = progress_text_placeholder
        self._width = width

    def set_value(self, value: NumberType) -> None:
        assert isinstance(value, NumberInstance), 'progress value must be numeric'
        assert 0 <= value <= 100, 'value must be between 0 and 100'
        self._progress = value
        self._render()

    def scale(self, *args, **kwargs) -> 'ProgressBar':
        raise WidgetTransformationNotImplemented()

    def resize(self, *args, **kwargs) -> 'ProgressBar':
        raise WidgetTransformationNotImplemented()

    def set_max_width(self, *args, **kwargs) -> 'ProgressBar':
        raise WidgetTransformationNotImplemented()

    def set_max_height(self, *args, **kwargs) -> 'ProgressBar':
        raise WidgetTransformationNotImplemented()

    def rotate(self, *args, **kwargs) -> 'ProgressBar':
        raise WidgetTransformationNotImplemented()

    def flip(self, *args, **kwargs) -> 'ProgressBar':
        raise WidgetTransformationNotImplemented()

    def get_value(self) -> NumberType:
        return self._progress

    def _apply_font(self) -> None:
        if self._progress_text_font is None:
            self._progress_text_font = self._font_name
        self._progress_text_font_height = int(self._font_size * self._progress_text_font_height_factor)
        self._progress_font = pygame_menu.font.get_font(
            self._progress_text_font, self._progress_text_font_height
        )
        self._box_height = self._font_render_string('TEST').get_height()

    def _draw(self, surface: 'pygame.Surface') -> None:
        # Draw title
        surface.blit(self._surface, (self._rect.x, self._rect.y))

        # Draw box
        box_rect = self._box.get_rect()
        box_rect.x += self._rect.x + self._box_margin[0] + self._box_pos
        box_rect.y += self._rect.y + self._box_margin[1]
        surface.blit(self._box, box_rect)

        # Draw box border
        if self._box_border_width > 0:
            pygame.draw.rect(surface, self._box_border_color, box_rect, self._box_border_width)

    def _render(self) -> Optional[bool]:
        if not hasattr(self, '_progress_font'):
            return False

        if not self._render_hash_changed(self._selected, self._title, self._visible, self.readonly, self._progress):
            return True

        # Create basic title
        self._surface = self._render_string(self._title, self.get_font_color_status())
        self._rect.width, self._rect.height = self._surface.get_size()

        # Create box
        self._box = make_surface(self._width, self._box_height, fill_color=self._box_background_color)
        box_progress = make_surface(int(self._width * self._progress / 100),
                                    self._box_height - self._box_progress_padding[0] - self._box_progress_padding[2],
                                    fill_color=self._box_progress_color)
        self._box.blit(box_progress, (self._box_progress_padding[1], self._box_progress_padding[0]))
        self._box_pos = self._rect.width

        # Create progress text
        text = self._progress_font.render(
            self._progress_text_placeholder.format(self._progress_text_format(self._progress)),
            self._font_antialias,
            self._progress_text_font_color
        )
        text_y = int((self._box_height - text.get_height()) / 2)
        if self._progress_text_align == ALIGN_LEFT:
            text_x = self._box_progress_padding[1]
        elif self._progress_text_align == ALIGN_CENTER:
            text_x = int((self._width - text.get_width()) / 2)
        else:
            text_x = self._width - self._box_progress_padding[3] - text.get_width()
        text_x += self._progress_text_margin[0]
        text_y += self._progress_text_margin[1]

        if self._progress_text_enabled:
            self._box.blit(text, (text_x, text_y))

        # Update maximum rect height
        self._rect.height = max(self._rect.height, self._box.get_height())
        self._rect.width += self._width + self._box_margin[0]

        # Finals
        self.force_menu_surface_update()

    def update(self, events: EventVectorType) -> bool:
        self.apply_update_callbacks(events)
        return False


class ProgressBarManager(AbstractWidgetManager, ABC):
    """
    ProgressBar manager.
    """

    def progress_bar(
        self,
        title: Any,
        default: NumberType = 0,
        onselect: CallbackType = None,
        progressbar_id: str = '',
        progress_text_format: ProgressBarTextFormatType = lambda x: str(round(x, 1)),
        selectable: bool = False,
        width: int = 150,
        **kwargs
    ) -> 'pygame_menu.widgets.ProgressBar':
        """
        Add a progress bar, which offers a bar that accepts a percentage from
        ``0`` to ``100``.

        If ``onselect`` is defined, the callback is executed as follows, where
        ``selected`` is a boolean representing the selected status:

        .. code-block:: python

            onselect(selected, widget, menu)

        kwargs (Optional)
            - ``align``                         (str) – Widget `alignment <https://pygame-menu.readthedocs.io/en/latest/_source/themes.html#alignment>`_
            - ``background_color``              (tuple, list, str, int, :py:class:`pygame.Color`, :py:class:`pygame_menu.baseimage.BaseImage`) – Color of the background. ``None`` for no-color
            - ``background_inflate``            (tuple, list) – Inflate background on x-axis and y-axis (x, y) in px
            - ``border_color``                  (tuple, list, str, int, :py:class:`pygame.Color`) – Widget border color. ``None`` for no-color
            - ``border_inflate``                (tuple, list) – Widget border inflate on x-axis and y-axis (x, y) in px
            - ``border_position``               (str, tuple, list) – Widget border positioning. It can be a single position, or a tuple/list of positions. Only are accepted: north, south, east, and west. See :py:mod:`pygame_menu.locals`
            - ``border_width``                  (int) – Border width in px. If ``0`` disables the border
            - ``box_background_color``          (tuple, list, str, int, :py:class:`pygame.Color`) – Background color of the box
            - ``box_border_color``              (tuple, list, str, int, :py:class:`pygame.Color`) – Border color of the box
            - ``box_border_width``              (int) - Border width of the box in px
            - ``box_margin``                    (tuple, list) - Box margin on x-axis and y-axis (x, y) respect to the title of the widget in px
            - ``box_progress_color``            (tuple, list, str, int, :py:class:`pygame.Color`) – Box progress color
            - ``box_progress_padding``          (int, float, tuple, list) – Box progress padding
            - ``cursor``                        (int, :py:class:`pygame.cursors.Cursor`, None) – Cursor of the widget if the mouse is placed over
            - ``float``                         (bool) - If ``True`` the widget don't contribute width/height to the Menu widget positioning computation, and don't add one unit to the rows
            - ``float_origin_position``         (bool) - If ``True`` the widget position is set to the top-left position of the Menu if the widget is floating
            - ``font_background_color``         (tuple, list, str, int, :py:class:`pygame.Color`, None) – Widget font background color
            - ``font_color``                    (tuple, list, str, int, :py:class:`pygame.Color`) – Widget font color
            - ``font_name``                     (str, :py:class:`pathlib.Path`, :py:class:`pygame.font.Font`) – Widget font path
            - ``font_shadow_color``             (tuple, list, str, int, :py:class:`pygame.Color`) – Font shadow color
            - ``font_shadow_offset``            (int) – Font shadow offset in px
            - ``font_shadow_position``          (str) – Font shadow position, see locals for position
            - ``font_shadow``                   (bool) – Font shadow is enabled or disabled
            - ``font_size``                     (int) – Font size of the widget
            - ``margin``                        (tuple, list) – Widget (left, bottom) margin in px
            - ``padding``                       (int, float, tuple, list) – Widget padding according to CSS rules. General shape: (top, right, bottom, left)
            - ``progress_text_align``           (str) - Align of the progress text, can be CENTER, LEFT or RIGHT. See :py:mod:`pygame_menu.locals`
            - ``progress_text_enabled``         (bool) - Enables the progress text over box
            - ``progress_text_font_color``      (tuple, list, str, int, :py:class:`pygame.Color`) – Progress font color. If ``None`` uses the same as the widget font
            - ``progress_text_font_hfactor``    (int, float) - Height factor of the font height relative to the widget font height
            - ``progress_text_font``            (str, :py:class:`pathlib.Path`, :py:class:`pygame.font.Font`) – Progress font. If ``None`` uses the same as the widget font
            - ``progress_text_margin``          (tuple, list) - Margin of the progress box on x-axis and y-axis in px
            - ``progress_text_placeholder``     (str) - Placeholder of the progress text, which considers as format the output of ``progress_text_format``. ``"{0} %"`` by default
            - ``readonly_color``                (tuple, list, str, int, :py:class:`pygame.Color`) – Color of the widget if readonly mode
            - ``readonly_selected_color``       (tuple, list, str, int, :py:class:`pygame.Color`) – Color of the widget if readonly mode and is selected
            - ``selection_color``               (tuple, list, str, int, :py:class:`pygame.Color`) – Color of the selected widget; only affects the font color
            - ``selection_effect``              (:py:class:`pygame_menu.widgets.core.Selection`) – Widget selection effect
            - ``shadow_color``                  (tuple, list, str, int, :py:class:`pygame.Color`) – Color of the widget shadow
            - ``shadow_radius``                 (int) - Border radius of the shadow
            - ``shadow_type``                   (str) - Shadow type, it can be ``'rectangular'`` or ``'ellipse'``
            - ``shadow_width``                  (int) - Width of the shadow. If ``0`` the shadow is disabled
            - ``tab_size``                      (int) – Width of a tab character

        .. note::

            All theme-related optional kwargs use the default Menu theme if not
            defined.

        .. note::

            This is applied only to the base Menu (not the currently displayed,
            stored in ``_current`` pointer); for such behaviour apply to
            :py:meth:`pygame_menu.menu.Menu.get_current` object.

        .. warning::

            Be careful with kwargs collision. Consider that all optional documented
            kwargs keys are removed from the object.

        :param title: Title of the progress bar
        :param default: Default value of the progressbar, from ``0`` to ``100``
        :param onselect: Callback executed when selecting the widget
        :param progressbar_id: ID of the progress bar
        :param progress_text_format: Format function of the progress text, which considers as input the progress value (0-100)
        :param selectable: Progress bar accepts user selection
        :param width: Progress bar width in px
        :param kwargs: Optional keyword arguments
        :return: Widget object
        :rtype: :py:class:`pygame_menu.widgets.ProgressBar`
        """
        assert isinstance(selectable, bool)

        # Filter widget attributes to avoid passing them to the callbacks
        attributes = self._filter_widget_attributes(kwargs)

        box_background_color = kwargs.pop('box_background_color', self._theme.widget_box_background_color)
        box_border_color = kwargs.pop('box_border_color', self._theme.widget_box_border_color)
        box_border_width = kwargs.pop('box_border_width', self._theme.widget_box_border_width)
        box_margin = kwargs.pop('box_margin', self._theme.widget_box_margin)
        box_progress_color = kwargs.pop('box_progress_color', (53, 172, 78))
        progress_text_font_color = kwargs.pop('progress_text_font_color', self._theme.widget_font_color)

        widget = pygame_menu.widgets.ProgressBar(
            title=title,
            progressbar_id=progressbar_id,
            default=default,
            width=width,
            onselect=onselect,
            box_background_color=box_background_color,
            box_border_color=box_border_color,
            box_border_width=box_border_width,
            box_margin=box_margin,
            box_progress_color=box_progress_color,
            progress_text_font_color=progress_text_font_color,
            progress_text_format=progress_text_format,
            **kwargs
        )
        widget.is_selectable = selectable

        self._configure_widget(widget=widget, **attributes)
        self._append_widget(widget)

        return widget



"""
pygame-menu
https://github.com/ppizarror/pygame-menu

EXAMPLE - SIMPLE
Super simple example of pygame-menu usage, featuring a selector and a button.
"""

import pygame_menu
from pygame_menu.examples import create_example_window

from typing import Tuple, Any

surface = create_example_window('Example - Simple', (600, 400))


def set_difficulty(selected: Tuple, value: Any) -> None:
    """
    Set the difficulty of the game.
    """
    print(f'Set difficulty to {selected[0]} ({value})')


def start_the_game() -> None:
    """
    Function that starts a game. This is raised by the menu button,
    here menu can be disabled, etc.
    """
    global user_name
    print(f'{user_name.get_value()}, Do the job here!')


menu = pygame_menu.Menu(
    height=300,
    theme=pygame_menu.themes.THEME_BLUE,
    title='Welcome',
    width=400
)

user_name = menu.add.text_input('Name: ', default='John Doe', maxchar=10)
menu.add.selector('Difficulty: ', [('Hard', 1), ('Easy', 2)], onchange=set_difficulty)
menu.add.button('Play', start_the_game)
menu.add.button('Quit', pygame_menu.events.EXIT)

if __name__ == '__main__':
    menu.mainloop(surface)
